#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from neutronclient.v2_0 import client
from oslotest import base

from ceilometer import neutron_client


class TestNeutronClientLBaaSV2(base.BaseTestCase):

    def setUp(self):
        super(TestNeutronClientLBaaSV2, self).setUp()
        self.nc = neutron_client.Client()

    @staticmethod
    def fake_list_lbaas_pools():
        return {
            'pools': [{
                'lb_algorithm': 'ROUND_ROBIN',
                'protocol': 'HTTP',
                'description': 'simple pool',
                'admin_state_up': True,
                'tenant_id': '1a3e005cf9ce40308c900bcb08e5320c',
                'healthmonitor_id': None,
                'listeners': [{
                    'id': "35cb8516-1173-4035-8dae-0dae3453f37f"
                    }
                ],
                'members': [{
                    'id': 'fcf23bde-8cf9-4616-883f-208cebcbf858'}
                ],
                'id': '4c0a0a5f-cf8f-44b7-b912-957daa8ce5e5',
                'name': 'pool1'
                }]
            }

    @staticmethod
    def fake_list_lbaas_members():
        return {
            'members': [{
                'weight': 1,
                'admin_state_up': True,
                'subnet_id': '013d3059-87a4-45a5-91e9-d721068ae0b2',
                'tenant_id': '1a3e005cf9ce40308c900bcb08e5320c',
                'address': '10.0.0.8',
                'protocol_port': 80,
                'id': 'fcf23bde-8cf9-4616-883f-208cebcbf858'
                }]
            }

    @staticmethod
    def fake_list_lbaas_healthmonitors():
        return {
            'healthmonitors': [{
                'admin_state_up': True,
                'tenant_id': '6f3584d5754048a18e30685362b88411',
                'delay': 1,
                'expected_codes': '200,201,202',
                'max_retries': 5,
                'http_method': 'GET',
                'timeout': 1,
                'pools': [{
                    'id': '74aa2010-a59f-4d35-a436-60a6da882819'
                    }],
                'url_path': '/index.html',
                'type': 'HTTP',
                'id': '0a9ac99d-0a09-4b18-8499-a0796850279a'
                }]
            }

    @staticmethod
    def fake_show_listener():
        return {
            'listener': {
                'default_pool_id': None,
                'protocol': 'HTTP',
                'description': '',
                'admin_state_up': True,
                'loadbalancers': [{
                    'id': 'a9729389-6147-41a3-ab22-a24aed8692b2'
                    }],
                'tenant_id': '3e4d8bec50a845fcb09e03a4375c691d',
                'connection_limit': 100,
                'protocol_port': 80,
                'id': '35cb8516-1173-4035-8dae-0dae3453f37f',
                'name': ''
                }
            }

    @staticmethod
    def fake_retrieve_loadbalancer_status():
        return {
            'statuses': {
                'loadbalancer': {
                    'operating_status': 'ONLINE',
                    'provisioning_status': 'ACTIVE',
                    'listeners': [{
                        'id': '35cb8516-1173-4035-8dae-0dae3453f37f',
                        'operating_status': 'ONLINE',
                        'provisioning_status': 'ACTIVE',
                        'pools': [{
                            'id': '4c0a0a5f-cf8f-44b7-b912-957daa8ce5e5',
                            'operating_status': 'ONLINE',
                            'provisioning_status': 'ACTIVE',
                            'members': [{
                                'id': 'fcf23bde-8cf9-4616-883f-208cebcbf858',
                                'operating_status': 'ONLINE',
                                'provisioning_status': 'ACTIVE'
                                }],
                            'healthmonitor': {
                                'id': '785131d2-8f7b-4fee-a7e7-3196e11b4518',
                                'provisioning_status': 'ACTIVE'
                                }
                            }]
                        }]
                    }
                }
            }

    @staticmethod
    def fake_retrieve_loadbalancer_status_complex():
        return {
            'statuses': {
                'loadbalancer': {
                    'operating_status': 'ONLINE',
                    'provisioning_status': 'ACTIVE',
                    'listeners': [{
                        'id': '35cb8516-1173-4035-8dae-0dae3453f37f',
                        'operating_status': 'ONLINE',
                        'provisioning_status': 'ACTIVE',
                        'pools': [{
                            'id': '4c0a0a5f-cf8f-44b7-b912-957daa8ce5e5',
                            'operating_status': 'ONLINE',
                            'provisioning_status': 'ACTIVE',
                            'members': [{
                                'id': 'fcf23bde-8cf9-4616-883f-208cebcbf858',
                                'operating_status': 'ONLINE',
                                'provisioning_status': 'ACTIVE'
                                },
                                {
                                'id': 'fcf23bde-8cf9-4616-883f-208cebcbf969',
                                'operating_status': 'OFFLINE',
                                'provisioning_status': 'ACTIVE'
                                }],
                            'healthmonitor': {
                                'id': '785131d2-8f7b-4fee-a7e7-3196e11b4518',
                                'provisioning_status': 'ACTIVE'
                                }
                            },
                            {
                            'id': '4c0a0a5f-cf8f-44b7-b912-957daa8ce6f6',
                            'operating_status': 'OFFLINE',
                            'provisioning_status': 'ACTIVE',
                            'members': [{
                                'id': 'fcf23bde-8cf9-4616-883f-208cebcbfa7a',
                                'operating_status': 'ONLINE',
                                'provisioning_status': 'ACTIVE'
                                }],
                            'healthmonitor': {
                                'id': '785131d2-8f7b-4fee-a7e7-3196e11b4629',
                                'provisioning_status': 'ACTIVE'
                                }
                            }]
                        },
                        {
                        'id': '35cb8516-1173-4035-8dae-0dae3453f48e',
                        'operating_status': 'OFFLINE',
                        'provisioning_status': 'ACTIVE',
                        'pools': [{
                            'id': '4c0a0a5f-cf8f-44b7-b912-957daa8ce7g7',
                            'operating_status': 'ONLINE',
                            'provisioning_status': 'ACTIVE',
                            'members': [{
                                'id': 'fcf23bde-8cf9-4616-883f-208cebcbfb8b',
                                'operating_status': 'ONLINE',
                                'provisioning_status': 'ACTIVE'
                                }],
                            'healthmonitor': {
                                'id': '785131d2-8f7b-4fee-a7e7-3196e11b473a',
                                'provisioning_status': 'ACTIVE'
                                }
                            }]
                        }]
                    }
                }
            }

    @staticmethod
    def fake_list_lbaas_listeners():
        return {
            'listeners': [{
                'default_pool_id': None,
                'protocol': 'HTTP',
                'description': '',
                'admin_state_up': True,
                'loadbalancers': [{
                    'id': 'a9729389-6147-41a3-ab22-a24aed8692b2'
                    }],
                'tenant_id': '3e4d8bec50a845fcb09e03a4375c691d',
                'connection_limit': 100,
                'protocol_port': 80,
                'id': '35cb8516-1173-4035-8dae-0dae3453f37f',
                'name': 'listener_one'
                }]}

    @mock.patch.object(client.Client,
                       'list_lbaas_pools')
    @mock.patch.object(client.Client,
                       'show_listener')
    @mock.patch.object(neutron_client.Client,
                       '_retrieve_loadbalancer_status_tree')
    def test_list_pools_v2(self, mock_status, mock_show, mock_list):
        mock_status.return_value = self.fake_retrieve_loadbalancer_status()
        mock_show.return_value = self.fake_show_listener()
        mock_list.return_value = self.fake_list_lbaas_pools()
        pools = self.nc.list_pools_v2()
        self.assertEqual(1, len(pools))
        for pool in pools:
            self.assertEqual('ONLINE', pool['status'])
            self.assertEqual('ROUND_ROBIN', pool['lb_method'])

    @mock.patch.object(client.Client,
                       'list_lbaas_pools')
    @mock.patch.object(client.Client,
                       'list_lbaas_members')
    @mock.patch.object(client.Client,
                       'show_listener')
    @mock.patch.object(neutron_client.Client,
                       '_retrieve_loadbalancer_status_tree')
    def test_list_members_v2(self, mock_status, mock_show, mock_list_members,
                             mock_list_pools):
        mock_status.return_value = self.fake_retrieve_loadbalancer_status()
        mock_show.return_value = self.fake_show_listener()
        mock_list_pools.return_value = self.fake_list_lbaas_pools()
        mock_list_members.return_value = self.fake_list_lbaas_members()
        members = self.nc.list_members_v2()
        self.assertEqual(1, len(members))
        for member in members:
            self.assertEqual('ONLINE', member['status'])
            self.assertEqual('4c0a0a5f-cf8f-44b7-b912-957daa8ce5e5',
                             member['pool_id'])

    @mock.patch.object(client.Client,
                       'list_lbaas_healthmonitors')
    def test_list_health_monitors_v2(self, mock_list_healthmonitors):
        mock_list_healthmonitors.return_value = (
            self.fake_list_lbaas_healthmonitors())
        healthmonitors = self.nc.list_health_monitors_v2()
        self.assertEqual(1, len(healthmonitors))
        for healthmonitor in healthmonitors:
            self.assertEqual(5, healthmonitor['max_retries'])

    @mock.patch.object(neutron_client.Client,
                       '_retrieve_loadbalancer_status_tree')
    def test_get_member_status(self, mock_status):
        mock_status.return_value = (
            self.fake_retrieve_loadbalancer_status_complex())
        loadbalancer_id = '5b1b1b6e-cf8f-44b7-b912-957daa8ce5e5'
        listener_id = '35cb8516-1173-4035-8dae-0dae3453f37f'
        pool_id = '4c0a0a5f-cf8f-44b7-b912-957daa8ce5e5'
        parent_id = [listener_id, pool_id]
        result_status = self.nc._get_member_status(loadbalancer_id,
                                                   parent_id)
        expected_keys = ['fcf23bde-8cf9-4616-883f-208cebcbf858',
                         'fcf23bde-8cf9-4616-883f-208cebcbf969']
        excepted_status = {
            'fcf23bde-8cf9-4616-883f-208cebcbf858': 'ONLINE',
            'fcf23bde-8cf9-4616-883f-208cebcbf969': 'OFFLINE'}

        for key in result_status.keys():
            self.assertIn(key, expected_keys)
            self.assertEqual(excepted_status[key], result_status[key])

    @mock.patch.object(neutron_client.Client,
                       '_retrieve_loadbalancer_status_tree')
    def test_get_pool_status(self, mock_status):
        mock_status.return_value = (
            self.fake_retrieve_loadbalancer_status_complex())
        loadbalancer_id = '5b1b1b6e-cf8f-44b7-b912-957daa8ce5e5'
        parent_id = '35cb8516-1173-4035-8dae-0dae3453f37f'
        result_status = self.nc._get_pool_status(loadbalancer_id,
                                                 parent_id)
        expected_keys = ['4c0a0a5f-cf8f-44b7-b912-957daa8ce5e5',
                         '4c0a0a5f-cf8f-44b7-b912-957daa8ce6f6']
        excepted_status = {
            '4c0a0a5f-cf8f-44b7-b912-957daa8ce5e5': 'ONLINE',
            '4c0a0a5f-cf8f-44b7-b912-957daa8ce6f6': 'OFFLINE'}

        for key in result_status.keys():
            self.assertIn(key, expected_keys)
            self.assertEqual(excepted_status[key], result_status[key])

    @mock.patch.object(neutron_client.Client,
                       '_retrieve_loadbalancer_status_tree')
    def test_get_listener_status(self, mock_status):
        mock_status.return_value = (
            self.fake_retrieve_loadbalancer_status_complex())
        loadbalancer_id = '5b1b1b6e-cf8f-44b7-b912-957daa8ce5e5'
        result_status = self.nc._get_listener_status(loadbalancer_id)
        expected_keys = ['35cb8516-1173-4035-8dae-0dae3453f37f',
                         '35cb8516-1173-4035-8dae-0dae3453f48e']
        excepted_status = {
            '35cb8516-1173-4035-8dae-0dae3453f37f': 'ONLINE',
            '35cb8516-1173-4035-8dae-0dae3453f48e': 'OFFLINE'}

        for key in result_status.keys():
            self.assertIn(key, expected_keys)
            self.assertEqual(excepted_status[key], result_status[key])

    @mock.patch.object(client.Client,
                       'list_listeners')
    @mock.patch.object(neutron_client.Client,
                       '_retrieve_loadbalancer_status_tree')
    def test_list_listener(self, mock_status, mock_list_listeners):
        mock_list_listeners.return_value = (
            self.fake_list_lbaas_listeners())
        mock_status.return_value = (
            self.fake_retrieve_loadbalancer_status())
        listeners = self.nc.list_listener()
        expected_key = '35cb8516-1173-4035-8dae-0dae3453f37f'
        expected_status = 'ONLINE'
        self.assertEqual(1, len(listeners))
        self.assertEqual(expected_key, listeners[0]['id'])
        self.assertEqual(expected_status, listeners[0]['operating_status'])



#
# Copyright 2012 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log
import oslo_messaging

from pecan import hooks

from ceilometer.i18n import _LE
from ceilometer import messaging
from ceilometer import storage

LOG = log.getLogger(__name__)

cfg.CONF.import_opt('telemetry_driver', 'ceilometer.publisher.messaging',
                    group='publisher_notifier')


class ConfigHook(hooks.PecanHook):
    """Attach the configuration object to the request.

    That allows controllers to get it.
    """

    @staticmethod
    def before(state):
        state.request.cfg = cfg.CONF


class DBHook(hooks.PecanHook):

    def __init__(self):
        self.storage_connection = DBHook.get_connection('metering')
        self.event_storage_connection = DBHook.get_connection('event')

        if (not self.storage_connection
           and not self.event_storage_connection):
            raise Exception("Api failed to start. Failed to connect to "
                            "databases, purpose:  %s" %
                            ', '.join(['metering', 'event']))

    def before(self, state):
        state.request.storage_conn = self.storage_connection
        state.request.event_storage_conn = self.event_storage_connection

    @staticmethod
    def get_connection(purpose):
        try:
            return storage.get_connection_from_config(cfg.CONF, purpose)
        except Exception as err:
            params = {"purpose": purpose, "err": err}
            LOG.exception(_LE("Failed to connect to db, purpose %(purpose)s "
                              "retry later: %(err)s") % params)


class NotifierHook(hooks.PecanHook):
    """Create and attach a notifier to the request.

    Usually, samples will be push to notification bus by notifier when they
    are posted via /v2/meters/ API.
    """

    def __init__(self):
        transport = messaging.get_transport()
        self.notifier = oslo_messaging.Notifier(
            transport, driver=cfg.CONF.publisher_notifier.telemetry_driver,
            publisher_id="ceilometer.api")

    def before(self, state):
        state.request.notifier = self.notifier


class TranslationHook(hooks.PecanHook):

    def after(self, state):
        # After a request has been done, we need to see if
        # ClientSideError has added an error onto the response.
        # If it has we need to get it info the thread-safe WSGI
        # environ to be used by the ParsableErrorMiddleware.
        if hasattr(state.response, 'translatable_error'):
            state.request.environ['translatable_error'] = (
                state.response.translatable_error)



#
# Copyright 2012 New Dream Network, LLC (DreamHost)
# Copyright 2013 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Base classes for API tests."""
import os
import uuid
import warnings

import fixtures
import mock
from oslo_config import fixture as fixture_config
from oslotest import mockpatch
import six
from six.moves.urllib import parse as urlparse
import sqlalchemy
from testtools import testcase

from ceilometer import storage
from ceilometer.tests import base as test_base
try:
    from ceilometer.tests import mocks
except ImportError:
    mocks = None   # happybase module is not Python 3 compatible yet


class MongoDbManager(fixtures.Fixture):

    def __init__(self, url):
        self._url = url

    def setUp(self):
        super(MongoDbManager, self).setUp()
        with warnings.catch_warnings():
            warnings.filterwarnings(
                action='ignore',
                message='.*you must provide a username and password.*')
            try:
                self.connection = storage.get_connection(
                    self.url, 'ceilometer.metering.storage')
                self.event_connection = storage.get_connection(
                    self.url, 'ceilometer.event.storage')
            except storage.StorageBadVersion as e:
                raise testcase.TestSkipped(six.text_type(e))

    @property
    def url(self):
        return '%(url)s_%(db)s' % {
            'url': self._url,
            'db': uuid.uuid4().hex
        }


class SQLManager(fixtures.Fixture):
    def __init__(self, url):
        db_name = 'ceilometer_%s' % uuid.uuid4().hex
        engine = sqlalchemy.create_engine(url)
        conn = engine.connect()
        self._create_database(conn, db_name)
        conn.close()
        engine.dispose()
        parsed = list(urlparse.urlparse(url))
        parsed[2] = '/' + db_name
        self.url = urlparse.urlunparse(parsed)

    def setUp(self):
        super(SQLManager, self).setUp()
        self.connection = storage.get_connection(
            self.url, 'ceilometer.metering.storage')
        self.event_connection = storage.get_connection(
            self.url, 'ceilometer.event.storage')


class PgSQLManager(SQLManager):
    @staticmethod
    def _create_database(conn, db_name):
        conn.connection.set_isolation_level(0)
        conn.execute('CREATE DATABASE %s WITH TEMPLATE template0;' % db_name)
        conn.connection.set_isolation_level(1)


class MySQLManager(SQLManager):
    @staticmethod
    def _create_database(conn, db_name):
        conn.execute('CREATE DATABASE %s;' % db_name)


class ElasticSearchManager(fixtures.Fixture):
    def __init__(self, url):
        self.url = url

    def setUp(self):
        super(ElasticSearchManager, self).setUp()
        self.connection = storage.get_connection(
            'sqlite://', 'ceilometer.metering.storage')
        self.event_connection = storage.get_connection(
            self.url, 'ceilometer.event.storage')
        # prefix each test with unique index name
        self.event_connection.index_name = 'events_%s' % uuid.uuid4().hex
        # force index on write so data is queryable right away
        self.event_connection._refresh_on_write = True


class HBaseManager(fixtures.Fixture):
    def __init__(self, url):
        self._url = url

    def setUp(self):
        super(HBaseManager, self).setUp()
        self.connection = storage.get_connection(
            self.url, 'ceilometer.metering.storage')
        self.event_connection = storage.get_connection(
            self.url, 'ceilometer.event.storage')
        # Unique prefix for each test to keep data is distinguished because
        # all test data is stored in one table
        data_prefix = str(uuid.uuid4().hex)

        def table(conn, name):
            return mocks.MockHBaseTable(name, conn, data_prefix)

        # Mock only real HBase connection, MConnection "table" method
        # stays origin.
        mock.patch('happybase.Connection.table', new=table).start()
        # We shouldn't delete data and tables after each test,
        # because it last for too long.
        # All tests tables will be deleted in setup-test-env.sh
        mock.patch("happybase.Connection.disable_table",
                   new=mock.MagicMock()).start()
        mock.patch("happybase.Connection.delete_table",
                   new=mock.MagicMock()).start()
        mock.patch("happybase.Connection.create_table",
                   new=mock.MagicMock()).start()

    @property
    def url(self):
        return '%s?table_prefix=%s&table_prefix_separator=%s' % (
            self._url,
            os.getenv("CEILOMETER_TEST_HBASE_TABLE_PREFIX", "test"),
            os.getenv("CEILOMETER_TEST_HBASE_TABLE_PREFIX_SEPARATOR", "_")
        )


class SQLiteManager(fixtures.Fixture):

    def __init__(self, url):
        self.url = url

    def setUp(self):
        super(SQLiteManager, self).setUp()
        self.connection = storage.get_connection(
            self.url, 'ceilometer.metering.storage')
        self.event_connection = storage.get_connection(
            self.url, 'ceilometer.event.storage')


@six.add_metaclass(test_base.SkipNotImplementedMeta)
class TestBase(test_base.BaseTestCase):

    DRIVER_MANAGERS = {
        'mongodb': MongoDbManager,
        'mysql': MySQLManager,
        'postgresql': PgSQLManager,
        'db2': MongoDbManager,
        'sqlite': SQLiteManager,
        'es': ElasticSearchManager,
    }
    if mocks is not None:
        DRIVER_MANAGERS['hbase'] = HBaseManager

    def setUp(self):
        super(TestBase, self).setUp()
        db_url = os.environ.get('OVERTEST_URL', "sqlite://").replace(
            "mysql://", "mysql+pymysql://")

        engine = urlparse.urlparse(db_url).scheme
        # in case some drivers have additional specification, for example:
        # PyMySQL will have scheme mysql+pymysql
        engine = engine.split('+')[0]

        # NOTE(Alexei_987) Shortcut to skip expensive db setUp
        test_method = self._get_test_method()
        if (hasattr(test_method, '_run_with')
                and engine not in test_method._run_with):
            raise testcase.TestSkipped(
                'Test is not applicable for %s' % engine)

        self.CONF = self.useFixture(fixture_config.Config()).conf
        self.CONF([], project='ceilometer', validate_default_values=True)

        manager = self.DRIVER_MANAGERS.get(engine)
        if not manager:
            self.skipTest("missing driver manager: %s" % engine)

        self.db_manager = manager(db_url)

        self.useFixture(self.db_manager)

        self.conn = self.db_manager.connection
        self.conn.upgrade()

        self.event_conn = self.db_manager.event_connection
        self.event_conn.upgrade()

        self.useFixture(mockpatch.Patch('ceilometer.storage.get_connection',
                                        side_effect=self._get_connection))

        # Set a default location for the pipeline config file so the
        # tests work even if ceilometer is not installed globally on
        # the system.
        self.CONF.import_opt('pipeline_cfg_file', 'ceilometer.pipeline')
        self.CONF.set_override(
            'pipeline_cfg_file',
            self.path_get('etc/ceilometer/pipeline.yaml')
        )

    def tearDown(self):
        self.event_conn.clear()
        self.event_conn = None
        self.conn.clear()
        self.conn = None
        super(TestBase, self).tearDown()

    def _get_connection(self, url, namespace):
        if namespace == "ceilometer.event.storage":
            return self.event_conn
        return self.conn


def run_with(*drivers):
    """Used to mark tests that are only applicable for certain db driver.

    Skips test if driver is not available.
    """
    def decorator(test):
        if isinstance(test, type) and issubclass(test, TestBase):
            # Decorate all test methods
            for attr in dir(test):
                value = getattr(test, attr)
                if callable(value) and attr.startswith('test_'):
                    if six.PY3:
                        value._run_with = drivers
                    else:
                        value.__func__._run_with = drivers
        else:
            test._run_with = drivers
        return test
    return decorator



#
# Copyright 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import keyword
import math
import re

from oslo_log import log
import six

from ceilometer.i18n import _
from ceilometer import sample
from ceilometer import transformer

LOG = log.getLogger(__name__)


class ArithmeticTransformer(transformer.TransformerBase):
    """Multi meter arithmetic transformer.

    Transformer that performs arithmetic operations
    over one or more meters and/or their metadata.
    """

    grouping_keys = ['resource_id']

    meter_name_re = re.compile(r'\$\(([\w\.\-]+)\)')

    def __init__(self, target=None, **kwargs):
        super(ArithmeticTransformer, self).__init__(**kwargs)
        target = target or {}
        self.target = target
        self.expr = target.get('expr', '')
        self.expr_escaped, self.escaped_names = self.parse_expr(self.expr)
        self.required_meters = list(self.escaped_names.values())
        self.misconfigured = len(self.required_meters) == 0
        if not self.misconfigured:
            self.reference_meter = self.required_meters[0]
            # convert to set for more efficient contains operation
            self.required_meters = set(self.required_meters)
            self.cache = collections.defaultdict(dict)
            self.latest_timestamp = None
        else:
            LOG.warning(_('Arithmetic transformer must use at least one'
                        ' meter in expression \'%s\''), self.expr)

    def _update_cache(self, _sample):
        """Update the cache with the latest sample."""
        escaped_name = self.escaped_names.get(_sample.name, '')
        if escaped_name not in self.required_meters:
            return
        self.cache[_sample.resource_id][escaped_name] = _sample

    def _check_requirements(self, resource_id):
        """Check if all the required meters are available in the cache."""
        return len(self.cache[resource_id]) == len(self.required_meters)

    def _calculate(self, resource_id):
        """Evaluate the expression and return a new sample if successful."""
        ns_dict = dict((m, s.as_dict()) for m, s
                       in six.iteritems(self.cache[resource_id]))
        ns = transformer.Namespace(ns_dict)
        try:
            new_volume = eval(self.expr_escaped, {}, ns)
            if math.isnan(new_volume):
                raise ArithmeticError(_('Expression evaluated to '
                                        'a NaN value!'))

            reference_sample = self.cache[resource_id][self.reference_meter]
            return sample.Sample(
                name=self.target.get('name', reference_sample.name),
                unit=self.target.get('unit', reference_sample.unit),
                type=self.target.get('type', reference_sample.type),
                volume=float(new_volume),
                user_id=reference_sample.user_id,
                project_id=reference_sample.project_id,
                resource_id=reference_sample.resource_id,
                timestamp=self.latest_timestamp,
                resource_metadata=reference_sample.resource_metadata
            )
        except Exception as e:
            LOG.warning(_('Unable to evaluate expression %(expr)s: %(exc)s'),
                        {'expr': self.expr, 'exc': e})

    def handle_sample(self, context, _sample):
        self._update_cache(_sample)
        self.latest_timestamp = _sample.timestamp

    def flush(self, context):
        new_samples = []
        cache_clean_list = []
        if not self.misconfigured:
            for resource_id in self.cache:
                if self._check_requirements(resource_id):
                    new_samples.append(self._calculate(resource_id))
                    cache_clean_list.append(resource_id)
            for res_id in cache_clean_list:
                self.cache.pop(res_id)
        return new_samples

    @classmethod
    def parse_expr(cls, expr):
        """Transforms meter names in the expression into valid identifiers.

        :param expr: unescaped expression
        :return: A tuple of the escaped expression and a dict representing
                 the translation of meter names into Python identifiers
        """

        class Replacer(object):
            """Replaces matched meter names with escaped names.

            If the meter name is not followed by parameter access in the
            expression, it defaults to accessing the 'volume' parameter.
            """

            def __init__(self, original_expr):
                self.original_expr = original_expr
                self.escaped_map = {}

            def __call__(self, match):
                meter_name = match.group(1)
                escaped_name = self.escape(meter_name)
                self.escaped_map[meter_name] = escaped_name

                if (match.end(0) == len(self.original_expr) or
                        self.original_expr[match.end(0)] != '.'):
                    escaped_name += '.volume'
                return escaped_name

            @staticmethod
            def escape(name):
                has_dot = '.' in name
                if has_dot:
                    name = name.replace('.', '_')

                if has_dot or name.endswith('ESC') or name in keyword.kwlist:
                    name = "_" + name + '_ESC'
                return name

        replacer = Replacer(expr)
        expr = re.sub(cls.meter_name_re, replacer, expr)
        return expr, replacer.escaped_map



#
# Copyright 2012 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Tests for ceilometer/storage/impl_log.py
"""
from oslotest import base

from ceilometer.storage import impl_log


class ConnectionTest(base.BaseTestCase):
    @staticmethod
    def test_get_connection():
        conn = impl_log.Connection(None)
        conn.record_metering_data({'counter_name': 'test',
                                   'resource_id': __name__,
                                   'counter_volume': 1,
                                   })



# Copyright 2014-2015 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log

from ceilometer.agent import plugin_base as plugin
from ceilometer.i18n import _LW
from ceilometer import keystone_client

LOG = log.getLogger(__name__)

cfg.CONF.import_group('service_credentials', 'ceilometer.keystone_client')


class EndpointDiscovery(plugin.DiscoveryBase):
    """Discovery that supplies service endpoints.

    This discovery should be used when the relevant APIs are not well suited
    to dividing the pollster's work into smaller pieces than a whole service
    at once. Example of this is the floating_ip pollster which calls
    nova.floating_ips.list() and therefore gets all floating IPs at once.
    """

    @staticmethod
    def discover(manager, param=None):
        endpoints = keystone_client.get_service_catalog(
            manager.keystone).get_urls(
                service_type=param,
                interface=cfg.CONF.service_credentials.interface,
                region_name=cfg.CONF.service_credentials.region_name)
        if not endpoints:
            LOG.warning(_LW('No endpoints found for service %s'),
                        "<all services>" if param is None else param)
            return []
        return endpoints



#
# Copyright 2012 eNovance <licensing@enovance.com>
# Copyright 2012 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log

import ceilometer
from ceilometer.compute import pollsters
from ceilometer.compute.pollsters import util
from ceilometer.compute.virt import inspector as virt_inspector
from ceilometer.i18n import _
from ceilometer import sample

LOG = log.getLogger(__name__)


class CPUPollster(pollsters.BaseComputePollster):

    def get_samples(self, manager, cache, resources):
        for instance in resources:
            LOG.debug('checking instance %s', instance.id)
            try:
                cpu_info = self.inspector.inspect_cpus(instance)
                LOG.debug("CPUTIME USAGE: %(instance)s %(time)d",
                          {'instance': instance,
                           'time': cpu_info.time})
                cpu_num = {'cpu_number': cpu_info.number}
                yield util.make_sample_from_instance(
                    instance,
                    name='cpu',
                    type=sample.TYPE_CUMULATIVE,
                    unit='ns',
                    volume=cpu_info.time,
                    additional_metadata=cpu_num,
                )
            except virt_inspector.InstanceNotFoundException as err:
                # Instance was deleted while getting samples. Ignore it.
                LOG.debug('Exception while getting samples %s', err)
            except virt_inspector.InstanceShutOffException as e:
                LOG.debug('Instance %(instance_id)s was shut off while '
                          'getting samples of %(pollster)s: %(exc)s',
                          {'instance_id': instance.id,
                           'pollster': self.__class__.__name__, 'exc': e})
            except ceilometer.NotImplementedError:
                # Selected inspector does not implement this pollster.
                LOG.debug('Obtaining CPU time is not implemented for %s',
                          self.inspector.__class__.__name__)
            except Exception as err:
                LOG.exception(_('could not get CPU time for %(id)s: %(e)s'),
                              {'id': instance.id, 'e': err})


class CPUUtilPollster(pollsters.BaseComputePollster):

    def get_samples(self, manager, cache, resources):
        self._inspection_duration = self._record_poll_time()
        for instance in resources:
            LOG.debug('Checking CPU util for instance %s', instance.id)
            try:
                cpu_info = self.inspector.inspect_cpu_util(
                    instance, self._inspection_duration)
                LOG.debug("CPU UTIL: %(instance)s %(util)d",
                          {'instance': instance,
                           'util': cpu_info.util})
                yield util.make_sample_from_instance(
                    instance,
                    name='cpu_util',
                    type=sample.TYPE_GAUGE,
                    unit='%',
                    volume=cpu_info.util,
                )
            except virt_inspector.InstanceNotFoundException as err:
                # Instance was deleted while getting samples. Ignore it.
                LOG.debug('Exception while getting samples %s', err)
            except ceilometer.NotImplementedError:
                # Selected inspector does not implement this pollster.
                LOG.debug('Obtaining CPU Util is not implemented for %s',
                          self.inspector.__class__.__name__)
            except Exception as err:
                LOG.exception(_('Could not get CPU Util for %(id)s: %(e)s'),
                              {'id': instance.id, 'e': err})



# Copyright 2014 Intel Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from oslo_config import cfg

from ceilometer.ipmi.pollsters import node
from ceilometer.tests.unit.ipmi.pollsters import base

CONF = cfg.CONF
CONF.import_opt('host', 'ceilometer.service')


class TestPowerPollster(base.TestPollsterBase):

    def fake_data(self):
        # data after parsing Intel Node Manager output
        return {"Current_value": ['13', '00']}

    def make_pollster(self):
        return node.PowerPollster()

    @mock.patch('ceilometer.pipeline.setup_pipeline', mock.MagicMock())
    def test_get_samples(self):
        self._test_get_samples()

        # only one sample, and value is 19(0x13 as current_value)
        self._verify_metering(1, 19, CONF.host)


class TestInletTemperaturePollster(base.TestPollsterBase):

    def fake_data(self):
        # data after parsing Intel Node Manager output
        return {"Current_value": ['23', '00']}

    def make_pollster(self):
        return node.InletTemperaturePollster()

    @mock.patch('ceilometer.pipeline.setup_pipeline', mock.MagicMock())
    def test_get_samples(self):
        self._test_get_samples()

        # only one sample, and value is 35(0x23 as current_value)
        self._verify_metering(1, 35, CONF.host)


class TestOutletTemperaturePollster(base.TestPollsterBase):

    def fake_data(self):
        # data after parsing Intel Node Manager output
        return {"Current_value": ['25', '00']}

    def make_pollster(self):
        return node.OutletTemperaturePollster()

    @mock.patch('ceilometer.pipeline.setup_pipeline', mock.MagicMock())
    def test_get_samples(self):
        self._test_get_samples()

        # only one sample, and value is 37(0x25 as current_value)
        self._verify_metering(1, 37, CONF.host)


class TestAirflowPollster(base.TestPollsterBase):

    def fake_data(self):
        # data after parsing Intel Node Manager output
        return {"Current_value": ['be', '00']}

    def make_pollster(self):
        return node.AirflowPollster()

    @mock.patch('ceilometer.pipeline.setup_pipeline', mock.MagicMock())
    def test_get_samples(self):
        self._test_get_samples()

        # only one sample, and value is 190(0xbe as current_value)
        self._verify_metering(1, 190, CONF.host)


class TestCUPSIndexPollster(base.TestPollsterBase):

    def fake_data(self):
        # data after parsing Intel Node Manager output
        return {"CUPS_Index": ['2e', '00']}

    def make_pollster(self):
        return node.CUPSIndexPollster()

    @mock.patch('ceilometer.pipeline.setup_pipeline', mock.MagicMock())
    def test_get_samples(self):
        self._test_get_samples()

        # only one sample, and value is 190(0xbe)
        self._verify_metering(1, 46, CONF.host)


class CPUUtilPollster(base.TestPollsterBase):

    def fake_data(self):
        # data after parsing Intel Node Manager output
        return {"CPU_Utilization":
                ['33', '00', '00', '00', '00', '00', '00', '00']}

    def make_pollster(self):
        return node.CPUUtilPollster()

    @mock.patch('ceilometer.pipeline.setup_pipeline', mock.MagicMock())
    def test_get_samples(self):
        self._test_get_samples()

        # only one sample, and value is 190(0xbe)
        self._verify_metering(1, 51, CONF.host)


class MemUtilPollster(base.TestPollsterBase):

    def fake_data(self):
        # data after parsing Intel Node Manager output
        return {"Mem_Utilization":
                ['05', '00', '00', '00', '00', '00', '00', '00']}

    def make_pollster(self):
        return node.MemUtilPollster()

    @mock.patch('ceilometer.pipeline.setup_pipeline', mock.MagicMock())
    def test_get_samples(self):
        self._test_get_samples()

        # only one sample, and value is 5(0x05)
        self._verify_metering(1, 5, CONF.host)


class IOUtilPollster(base.TestPollsterBase):

    def fake_data(self):
        # data after parsing Intel Node Manager output
        return {"IO_Utilization":
                ['00', '00', '00', '00', '00', '00', '00', '00']}

    def make_pollster(self):
        return node.IOUtilPollster()

    @mock.patch('ceilometer.pipeline.setup_pipeline', mock.MagicMock())
    def test_get_samples(self):
        self._test_get_samples()

        # only one sample, and value is 0(0x00)
        self._verify_metering(1, 0, CONF.host)



#
# Copyright 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Sample data for test_ipmi.

This data is provided as a sample of the data expected from the ipmitool
driver in the Ironic project, which is the publisher of the notifications
being tested.
"""


TEMPERATURE_DATA = {
    'DIMM GH VR Temp (0x3b)': {
        'Status': 'ok',
        'Deassertions Enabled': 'unc+ ucr+ unr+',
        'Sensor Reading': '26 (+/- 0.500) degrees C',
        'Entity ID': '20.6 (Power Module)',
        'Assertions Enabled': 'unc+ ucr+ unr+',
        'Positive Hysteresis': '4.000',
        'Assertion Events': '',
        'Upper non-critical': '95.000',
        'Event Message Control': 'Per-threshold',
        'Upper non-recoverable': '105.000',
        'Normal Maximum': '112.000',
        'Maximum sensor range': 'Unspecified',
        'Sensor Type (Analog)': 'Temperature',
        'Readable Thresholds': 'unc ucr unr',
        'Negative Hysteresis': 'Unspecified',
        'Threshold Read Mask': 'unc ucr unr',
        'Upper critical': '100.000',
        'Sensor ID': 'DIMM GH VR Temp (0x3b)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '16.000'
    },
    'CPU1 VR Temp (0x36)': {
        'Status': 'ok',
        'Deassertions Enabled': 'unc+ ucr+ unr+',
        'Sensor Reading': '32 (+/- 0.500) degrees C',
        'Entity ID': '20.1 (Power Module)',
        'Assertions Enabled': 'unc+ ucr+ unr+',
        'Positive Hysteresis': '4.000',
        'Assertion Events': '',
        'Upper non-critical': '95.000',
        'Event Message Control': 'Per-threshold',
        'Upper non-recoverable': '105.000',
        'Normal Maximum': '112.000',
        'Maximum sensor range': 'Unspecified',
        'Sensor Type (Analog)': 'Temperature',
        'Readable Thresholds': 'unc ucr unr',
        'Negative Hysteresis': 'Unspecified',
        'Threshold Read Mask': 'unc ucr unr',
        'Upper critical': '100.000',
        'Sensor ID': 'CPU1 VR Temp (0x36)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '16.000'
    },
    'DIMM EF VR Temp (0x3a)': {
        'Status': 'ok',
        'Deassertions Enabled': 'unc+ ucr+ unr+',
        'Sensor Reading': '26 (+/- 0.500) degrees C',
        'Entity ID': '20.5 (Power Module)',
        'Assertions Enabled': 'unc+ ucr+ unr+',
        'Positive Hysteresis': '4.000',
        'Assertion Events': '',
        'Upper non-critical': '95.000',
        'Event Message Control': 'Per-threshold',
        'Upper non-recoverable': '105.000',
        'Normal Maximum': '112.000',
        'Maximum sensor range': 'Unspecified',
        'Sensor Type (Analog)': 'Temperature',
        'Readable Thresholds': 'unc ucr unr',
        'Negative Hysteresis': 'Unspecified',
        'Threshold Read Mask': 'unc ucr unr',
        'Upper critical': '100.000',
        'Sensor ID': 'DIMM EF VR Temp (0x3a)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '16.000'
    },
    'CPU2 VR Temp (0x37)': {
        'Status': 'ok',
        'Deassertions Enabled': 'unc+ ucr+ unr+',
        'Sensor Reading': '31 (+/- 0.500) degrees C',
        'Entity ID': '20.2 (Power Module)',
        'Assertions Enabled': 'unc+ ucr+ unr+',
        'Positive Hysteresis': '4.000',
        'Assertion Events': '',
        'Upper non-critical': '95.000',
        'Event Message Control': 'Per-threshold',
        'Upper non-recoverable': '105.000',
        'Normal Maximum': '112.000',
        'Maximum sensor range': 'Unspecified',
        'Sensor Type (Analog)': 'Temperature',
        'Readable Thresholds': 'unc ucr unr',
        'Negative Hysteresis': 'Unspecified',
        'Threshold Read Mask': 'unc ucr unr',
        'Upper critical': '100.000',
        'Sensor ID': 'CPU2 VR Temp (0x37)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '16.000'
    },
    'Ambient Temp (0x32)': {
        'Status': 'ok',
        'Sensor Reading': '25 (+/- 0) degrees C',
        'Entity ID': '12.1 (Front Panel Board)',
        'Assertions Enabled': 'unc+ ucr+ unr+',
        'Event Message Control': 'Per-threshold',
        'Assertion Events': '',
        'Upper non-critical': '43.000',
        'Deassertions Enabled': 'unc+ ucr+ unr+',
        'Upper non-recoverable': '50.000',
        'Positive Hysteresis': '4.000',
        'Maximum sensor range': 'Unspecified',
        'Sensor Type (Analog)': 'Temperature',
        'Readable Thresholds': 'unc ucr unr',
        'Negative Hysteresis': 'Unspecified',
        'Threshold Read Mask': 'unc ucr unr',
        'Upper critical': '46.000',
        'Sensor ID': 'Ambient Temp (0x32)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '25.000'
    },
    'Mezz Card Temp (0x35)': {
        'Status': 'Disabled',
        'Sensor Reading': 'Disabled',
        'Entity ID': '44.1 (I/O Module)',
        'Event Message Control': 'Per-threshold',
        'Upper non-critical': '70.000',
        'Upper non-recoverable': '85.000',
        'Positive Hysteresis': '4.000',
        'Maximum sensor range': 'Unspecified',
        'Sensor Type (Analog)': 'Temperature',
        'Readable Thresholds': 'unc ucr unr',
        'Negative Hysteresis': 'Unspecified',
        'Threshold Read Mask': 'unc ucr unr',
        'Upper critical': '80.000',
        'Sensor ID': 'Mezz Card Temp (0x35)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '25.000'
    },
    'PCH Temp (0x3c)': {
        'Status': 'ok',
        'Deassertions Enabled': 'unc+ ucr+ unr+',
        'Sensor Reading': '46 (+/- 0.500) degrees C',
        'Entity ID': '45.1 (Processor/IO Module)',
        'Assertions Enabled': 'unc+ ucr+ unr+',
        'Positive Hysteresis': '4.000',
        'Assertion Events': '',
        'Upper non-critical': '93.000',
        'Event Message Control': 'Per-threshold',
        'Upper non-recoverable': '103.000',
        'Normal Maximum': '112.000',
        'Maximum sensor range': 'Unspecified',
        'Sensor Type (Analog)': 'Temperature',
        'Readable Thresholds': 'unc ucr unr',
        'Negative Hysteresis': 'Unspecified',
        'Threshold Read Mask': 'unc ucr unr',
        'Upper critical': '98.000',
        'Sensor ID': 'PCH Temp (0x3c)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '16.000'
    },
    'DIMM CD VR Temp (0x39)': {
        'Status': 'ok',
        'Deassertions Enabled': 'unc+ ucr+ unr+',
        'Sensor Reading': '27 (+/- 0.500) degrees C',
        'Entity ID': '20.4 (Power Module)',
        'Assertions Enabled': 'unc+ ucr+ unr+',
        'Positive Hysteresis': '4.000',
        'Assertion Events': '',
        'Upper non-critical': '95.000',
        'Event Message Control': 'Per-threshold',
        'Upper non-recoverable': '105.000',
        'Normal Maximum': '112.000',
        'Maximum sensor range': 'Unspecified',
        'Sensor Type (Analog)': 'Temperature',
        'Readable Thresholds': 'unc ucr unr',
        'Negative Hysteresis': 'Unspecified',
        'Threshold Read Mask': 'unc ucr unr',
        'Upper critical': '100.000',
        'Sensor ID': 'DIMM CD VR Temp (0x39)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '16.000'
    },
    'PCI Riser 2 Temp (0x34)': {
        'Status': 'ok',
        'Deassertions Enabled': 'unc+ ucr+ unr+',
        'Sensor Reading': '30 (+/- 0) degrees C',
        'Entity ID': '16.2 (System Internal Expansion Board)',
        'Assertions Enabled': 'unc+ ucr+ unr+',
        'Positive Hysteresis': '4.000',
        'Assertion Events': '',
        'Upper non-critical': '70.000',
        'Event Message Control': 'Per-threshold',
        'Upper non-recoverable': '85.000',
        'Normal Maximum': '112.000',
        'Maximum sensor range': 'Unspecified',
        'Sensor Type (Analog)': 'Temperature',
        'Readable Thresholds': 'unc ucr unr',
        'Negative Hysteresis': 'Unspecified',
        'Threshold Read Mask': 'unc ucr unr',
        'Upper critical': '80.000',
        'Sensor ID': 'PCI Riser 2 Temp (0x34)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '16.000'
    },
    'DIMM AB VR Temp (0x38)': {
        'Status': 'ok',
        'Deassertions Enabled': 'unc+ ucr+ unr+',
        'Sensor Reading': '28 (+/- 0.500) degrees C',
        'Entity ID': '20.3 (Power Module)',
        'Assertions Enabled': 'unc+ ucr+ unr+',
        'Positive Hysteresis': '4.000',
        'Assertion Events': '',
        'Upper non-critical': '95.000',
        'Event Message Control': 'Per-threshold',
        'Upper non-recoverable': '105.000',
        'Normal Maximum': '112.000',
        'Maximum sensor range': 'Unspecified',
        'Sensor Type (Analog)': 'Temperature',
        'Readable Thresholds': 'unc ucr unr',
        'Negative Hysteresis': 'Unspecified',
        'Threshold Read Mask': 'unc ucr unr',
        'Upper critical': '100.000',
        'Sensor ID': 'DIMM AB VR Temp (0x38)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '16.000'
    },
    'PCI Riser 1 Temp (0x33)': {
        'Status': 'ok',
        'Deassertions Enabled': 'unc+ ucr+ unr+',
        'Sensor Reading': '38 (+/- 0) degrees C',
        'Entity ID': '16.1 (System Internal Expansion Board)',
        'Assertions Enabled': 'unc+ ucr+ unr+',
        'Positive Hysteresis': '4.000',
        'Assertion Events': '',
        'Upper non-critical': '70.000',
        'Event Message Control': 'Per-threshold',
        'Upper non-recoverable': '85.000',
        'Normal Maximum': '112.000',
        'Maximum sensor range': 'Unspecified',
        'Sensor Type (Analog)': 'Temperature',
        'Readable Thresholds': 'unc ucr unr',
        'Negative Hysteresis': 'Unspecified',
        'Threshold Read Mask': 'unc ucr unr',
        'Upper critical': '80.000',
        'Sensor ID': 'PCI Riser 1 Temp (0x33)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '16.000'
    },
}


CURRENT_DATA = {
    'Avg Power (0x2e)': {
        'Status': 'ok',
        'Sensor Reading': '130 (+/- 0) Watts',
        'Entity ID': '21.0 (Power Management)',
        'Assertions Enabled': '',
        'Event Message Control': 'Per-threshold',
        'Readable Thresholds': 'No Thresholds',
        'Positive Hysteresis': 'Unspecified',
        'Sensor Type (Analog)': 'Current',
        'Negative Hysteresis': 'Unspecified',
        'Maximum sensor range': 'Unspecified',
        'Sensor ID': 'Avg Power (0x2e)',
        'Assertion Events': '',
        'Minimum sensor range': '2550.000',
        'Settable Thresholds': 'No Thresholds'
    }
}


FAN_DATA = {
    'Fan 4A Tach (0x46)': {
        'Status': 'ok',
        'Sensor Reading': '6900 (+/- 0) RPM',
        'Entity ID': '29.4 (Fan Device)',
        'Assertions Enabled': 'lcr-',
        'Normal Minimum': '2580.000',
        'Positive Hysteresis': '120.000',
        'Assertion Events': '',
        'Event Message Control': 'Per-threshold',
        'Normal Maximum': '15300.000',
        'Deassertions Enabled': 'lcr-',
        'Sensor Type (Analog)': 'Fan',
        'Lower critical': '1920.000',
        'Negative Hysteresis': '120.000',
        'Threshold Read Mask': 'lcr',
        'Maximum sensor range': 'Unspecified',
        'Readable Thresholds': 'lcr',
        'Sensor ID': 'Fan 4A Tach (0x46)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '4020.000'
    },
    'Fan 5A Tach (0x48)': {
        'Status': 'ok',
        'Sensor Reading': '7140 (+/- 0) RPM',
        'Entity ID': '29.5 (Fan Device)',
        'Assertions Enabled': 'lcr-',
        'Normal Minimum': '2580.000',
        'Positive Hysteresis': '120.000',
        'Assertion Events': '',
        'Event Message Control': 'Per-threshold',
        'Normal Maximum': '15300.000',
        'Deassertions Enabled': 'lcr-',
        'Sensor Type (Analog)': 'Fan',
        'Lower critical': '1920.000',
        'Negative Hysteresis': '120.000',
        'Threshold Read Mask': 'lcr',
        'Maximum sensor range': 'Unspecified',
        'Readable Thresholds': 'lcr',
        'Sensor ID': 'Fan 5A Tach (0x48)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '4020.000'
    },
    'Fan 3A Tach (0x44)': {
        'Status': 'ok',
        'Sensor Reading': '6900 (+/- 0) RPM',
        'Entity ID': '29.3 (Fan Device)',
        'Assertions Enabled': 'lcr-',
        'Normal Minimum': '2580.000',
        'Positive Hysteresis': '120.000',
        'Assertion Events': '',
        'Event Message Control': 'Per-threshold',
        'Normal Maximum': '15300.000',
        'Deassertions Enabled': 'lcr-',
        'Sensor Type (Analog)': 'Fan',
        'Lower critical': '1920.000',
        'Negative Hysteresis': '120.000',
        'Threshold Read Mask': 'lcr',
        'Maximum sensor range': 'Unspecified',
        'Readable Thresholds': 'lcr',
        'Sensor ID': 'Fan 3A Tach (0x44)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '4020.000'
    },
    'Fan 1A Tach (0x40)': {
        'Status': 'ok',
        'Sensor Reading': '6960 (+/- 0) RPM',
        'Entity ID': '29.1 (Fan Device)',
        'Assertions Enabled': 'lcr-',
        'Normal Minimum': '2580.000',
        'Positive Hysteresis': '120.000',
        'Assertion Events': '',
        'Event Message Control': 'Per-threshold',
        'Normal Maximum': '15300.000',
        'Deassertions Enabled': 'lcr-',
        'Sensor Type (Analog)': 'Fan',
        'Lower critical': '1920.000',
        'Negative Hysteresis': '120.000',
        'Threshold Read Mask': 'lcr',
        'Maximum sensor range': 'Unspecified',
        'Readable Thresholds': 'lcr',
        'Sensor ID': 'Fan 1A Tach (0x40)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '4020.000'
    },
    'Fan 3B Tach (0x45)': {
        'Status': 'ok',
        'Sensor Reading': '7104 (+/- 0) RPM',
        'Entity ID': '29.3 (Fan Device)',
        'Assertions Enabled': 'lcr-',
        'Normal Minimum': '2752.000',
        'Positive Hysteresis': '128.000',
        'Assertion Events': '',
        'Event Message Control': 'Per-threshold',
        'Normal Maximum': '16320.000',
        'Deassertions Enabled': 'lcr-',
        'Sensor Type (Analog)': 'Fan',
        'Lower critical': '1920.000',
        'Negative Hysteresis': '128.000',
        'Threshold Read Mask': 'lcr',
        'Maximum sensor range': 'Unspecified',
        'Readable Thresholds': 'lcr',
        'Sensor ID': 'Fan 3B Tach (0x45)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '3968.000'
    },
    'Fan 2A Tach (0x42)': {
        'Status': 'ok',
        'Sensor Reading': '7080 (+/- 0) RPM',
        'Entity ID': '29.2 (Fan Device)',
        'Assertions Enabled': 'lcr-',
        'Normal Minimum': '2580.000',
        'Positive Hysteresis': '120.000',
        'Assertion Events': '',
        'Event Message Control': 'Per-threshold',
        'Normal Maximum': '15300.000',
        'Deassertions Enabled': 'lcr-',
        'Sensor Type (Analog)': 'Fan',
        'Lower critical': '1920.000',
        'Negative Hysteresis': '120.000',
        'Threshold Read Mask': 'lcr',
        'Maximum sensor range': 'Unspecified',
        'Readable Thresholds': 'lcr',
        'Sensor ID': 'Fan 2A Tach (0x42)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '4020.000'
    },
    'Fan 4B Tach (0x47)': {
        'Status': 'ok',
        'Sensor Reading': '7488 (+/- 0) RPM',
        'Entity ID': '29.4 (Fan Device)',
        'Assertions Enabled': 'lcr-',
        'Normal Minimum': '2752.000',
        'Positive Hysteresis': '128.000',
        'Assertion Events': '',
        'Event Message Control': 'Per-threshold',
        'Normal Maximum': '16320.000',
        'Deassertions Enabled': 'lcr-',
        'Sensor Type (Analog)': 'Fan',
        'Lower critical': '1920.000',
        'Negative Hysteresis': '128.000',
        'Threshold Read Mask': 'lcr',
        'Maximum sensor range': 'Unspecified',
        'Readable Thresholds': 'lcr',
        'Sensor ID': 'Fan 4B Tach (0x47)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '3968.000'
    },
    'Fan 2B Tach (0x43)': {
        'Status': 'ok',
        'Sensor Reading': '7168 (+/- 0) RPM',
        'Entity ID': '29.2 (Fan Device)',
        'Assertions Enabled': 'lcr-',
        'Normal Minimum': '2752.000',
        'Positive Hysteresis': '128.000',
        'Assertion Events': '',
        'Event Message Control': 'Per-threshold',
        'Normal Maximum': '16320.000',
        'Deassertions Enabled': 'lcr-',
        'Sensor Type (Analog)': 'Fan',
        'Lower critical': '1920.000',
        'Negative Hysteresis': '128.000',
        'Threshold Read Mask': 'lcr',
        'Maximum sensor range': 'Unspecified',
        'Readable Thresholds': 'lcr',
        'Sensor ID': 'Fan 2B Tach (0x43)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '3968.000'
    },
    'Fan 5B Tach (0x49)': {
        'Status': 'ok',
        'Sensor Reading': '7296 (+/- 0) RPM',
        'Entity ID': '29.5 (Fan Device)',
        'Assertions Enabled': 'lcr-',
        'Normal Minimum': '2752.000',
        'Positive Hysteresis': '128.000',
        'Assertion Events': '',
        'Event Message Control': 'Per-threshold',
        'Normal Maximum': '16320.000',
        'Deassertions Enabled': 'lcr-',
        'Sensor Type (Analog)': 'Fan',
        'Lower critical': '1920.000',
        'Negative Hysteresis': '128.000',
        'Threshold Read Mask': 'lcr',
        'Maximum sensor range': 'Unspecified',
        'Readable Thresholds': 'lcr',
        'Sensor ID': 'Fan 5B Tach (0x49)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '3968.000'
    },
    'Fan 1B Tach (0x41)': {
        'Status': 'ok',
        'Sensor Reading': '7296 (+/- 0) RPM',
        'Entity ID': '29.1 (Fan Device)',
        'Assertions Enabled': 'lcr-',
        'Normal Minimum': '2752.000',
        'Positive Hysteresis': '128.000',
        'Assertion Events': '',
        'Event Message Control': 'Per-threshold',
        'Normal Maximum': '16320.000',
        'Deassertions Enabled': 'lcr-',
        'Sensor Type (Analog)': 'Fan',
        'Lower critical': '1920.000',
        'Negative Hysteresis': '128.000',
        'Threshold Read Mask': 'lcr',
        'Maximum sensor range': 'Unspecified',
        'Readable Thresholds': 'lcr',
        'Sensor ID': 'Fan 1B Tach (0x41)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '3968.000'
    },
    'Fan 6B Tach (0x4b)': {
        'Status': 'ok',
        'Sensor Reading': '7616 (+/- 0) RPM',
        'Entity ID': '29.6 (Fan Device)',
        'Assertions Enabled': 'lcr-',
        'Normal Minimum': '2752.000',
        'Positive Hysteresis': '128.000',
        'Assertion Events': '',
        'Event Message Control': 'Per-threshold',
        'Normal Maximum': '16320.000',
        'Deassertions Enabled': 'lcr-',
        'Sensor Type (Analog)': 'Fan',
        'Lower critical': '1920.000',
        'Negative Hysteresis': '128.000',
        'Threshold Read Mask': 'lcr',
        'Maximum sensor range': 'Unspecified',
        'Readable Thresholds': 'lcr',
        'Sensor ID': 'Fan 6B Tach (0x4b)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '3968.000'
    },
    'Fan 6A Tach (0x4a)': {
        'Status': 'ok',
        'Sensor Reading': '7080 (+/- 0) RPM',
        'Entity ID': '29.6 (Fan Device)',
        'Assertions Enabled': 'lcr-',
        'Normal Minimum': '2580.000',
        'Positive Hysteresis': '120.000',
        'Assertion Events': '',
        'Event Message Control': 'Per-threshold',
        'Normal Maximum': '15300.000',
        'Deassertions Enabled': 'lcr-',
        'Sensor Type (Analog)': 'Fan',
        'Lower critical': '1920.000',
        'Negative Hysteresis': '120.000',
        'Threshold Read Mask': 'lcr',
        'Maximum sensor range': 'Unspecified',
        'Readable Thresholds': 'lcr',
        'Sensor ID': 'Fan 6A Tach (0x4a)',
        'Settable Thresholds': '',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '4020.000'
    }
}


VOLTAGE_DATA = {
    'Planar 12V (0x18)': {
        'Status': 'ok',
        'Sensor Reading': '12.312 (+/- 0) Volts',
        'Entity ID': '7.1 (System Board)',
        'Assertions Enabled': 'lcr- ucr+',
        'Event Message Control': 'Per-threshold',
        'Assertion Events': '',
        'Maximum sensor range': 'Unspecified',
        'Positive Hysteresis': '0.108',
        'Deassertions Enabled': 'lcr- ucr+',
        'Sensor Type (Analog)': 'Voltage',
        'Lower critical': '10.692',
        'Negative Hysteresis': '0.108',
        'Threshold Read Mask': 'lcr ucr',
        'Upper critical': '13.446',
        'Readable Thresholds': 'lcr ucr',
        'Sensor ID': 'Planar 12V (0x18)',
        'Settable Thresholds': 'lcr ucr',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '12.042'
    },
    'Planar 3.3V (0x16)': {
        'Status': 'ok',
        'Sensor Reading': '3.309 (+/- 0) Volts',
        'Entity ID': '7.1 (System Board)',
        'Assertions Enabled': 'lcr- ucr+',
        'Event Message Control': 'Per-threshold',
        'Assertion Events': '',
        'Maximum sensor range': 'Unspecified',
        'Positive Hysteresis': '0.028',
        'Deassertions Enabled': 'lcr- ucr+',
        'Sensor Type (Analog)': 'Voltage',
        'Lower critical': '3.039',
        'Negative Hysteresis': '0.028',
        'Threshold Read Mask': 'lcr ucr',
        'Upper critical': '3.564',
        'Readable Thresholds': 'lcr ucr',
        'Sensor ID': 'Planar 3.3V (0x16)',
        'Settable Thresholds': 'lcr ucr',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '3.309'
    },
    'Planar VBAT (0x1c)': {
        'Status': 'ok',
        'Sensor Reading': '3.137 (+/- 0) Volts',
        'Entity ID': '7.1 (System Board)',
        'Assertions Enabled': 'lnc- lcr-',
        'Event Message Control': 'Per-threshold',
        'Assertion Events': '',
        'Readable Thresholds': 'lcr lnc',
        'Positive Hysteresis': '0.025',
        'Deassertions Enabled': 'lnc- lcr-',
        'Sensor Type (Analog)': 'Voltage',
        'Lower critical': '2.095',
        'Negative Hysteresis': '0.025',
        'Lower non-critical': '2.248',
        'Maximum sensor range': 'Unspecified',
        'Sensor ID': 'Planar VBAT (0x1c)',
        'Settable Thresholds': 'lcr lnc',
        'Threshold Read Mask': 'lcr lnc',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '3.010'
    },
    'Planar 5V (0x17)': {
        'Status': 'ok',
        'Sensor Reading': '5.062 (+/- 0) Volts',
        'Entity ID': '7.1 (System Board)',
        'Assertions Enabled': 'lcr- ucr+',
        'Event Message Control': 'Per-threshold',
        'Assertion Events': '',
        'Maximum sensor range': 'Unspecified',
        'Positive Hysteresis': '0.045',
        'Deassertions Enabled': 'lcr- ucr+',
        'Sensor Type (Analog)': 'Voltage',
        'Lower critical': '4.475',
        'Negative Hysteresis': '0.045',
        'Threshold Read Mask': 'lcr ucr',
        'Upper critical': '5.582',
        'Readable Thresholds': 'lcr ucr',
        'Sensor ID': 'Planar 5V (0x17)',
        'Settable Thresholds': 'lcr ucr',
        'Minimum sensor range': 'Unspecified',
        'Nominal Reading': '4.995'
    }
}


SENSOR_DATA = {
    'message_id': 'f22188ca-c068-47ce-a3e5-0e27ffe234c6',
    'publisher_id': 'f23188ca-c068-47ce-a3e5-0e27ffe234c6',
    'payload': {
        'instance_uuid': 'f11251ax-c568-25ca-4582-0x27add644c6',
        'timestamp': '20140223134852',
        'node_uuid': 'f4982fd2-2f2b-4bb5-9aff-48aac801d1ad',
        'event_type': 'hardware.ipmi.metrics.update',
        'payload': {
            'Temperature': TEMPERATURE_DATA,
            'Current': CURRENT_DATA,
            'Fan': FAN_DATA,
            'Voltage': VOLTAGE_DATA
        }
    }
}


EMPTY_PAYLOAD = {
    'message_id': 'f22188ca-c068-47ce-a3e5-0e27ffe234c6',
    'publisher_id': 'f23188ca-c068-47ce-a3e5-0e27ffe234c6',
    'payload': {
        'instance_uuid': 'f11251ax-c568-25ca-4582-0x27add644c6',
        'timestamp': '20140223134852',
        'node_uuid': 'f4982fd2-2f2b-4bb5-9aff-48aac801d1ad',
        'event_type': 'hardware.ipmi.metrics.update',
        'payload': {
        }
    }
}


MISSING_SENSOR = {
    'message_id': 'f22188ca-c068-47ce-a3e5-0e27ffe234c6',
    'publisher_id': 'f23188ca-c068-47ce-a3e5-0e27ffe234c6',
    'payload': {
        'instance_uuid': 'f11251ax-c568-25ca-4582-0x27add644c6',
        'timestamp': '20140223134852',
        'node_uuid': 'f4982fd2-2f2b-4bb5-9aff-48aac801d1ad',
        'event_type': 'hardware.ipmi.metrics.update',
        'payload': {
            'Temperature': {
                'PCI Riser 1 Temp (0x33)': {
                    'Status': 'ok',
                    'Deassertions Enabled': 'unc+ ucr+ unr+',
                    'Entity ID': '16.1 (System Internal Expansion Board)',
                    'Assertions Enabled': 'unc+ ucr+ unr+',
                    'Positive Hysteresis': '4.000',
                    'Assertion Events': '',
                    'Upper non-critical': '70.000',
                    'Event Message Control': 'Per-threshold',
                    'Upper non-recoverable': '85.000',
                    'Normal Maximum': '112.000',
                    'Maximum sensor range': 'Unspecified',
                    'Sensor Type (Analog)': 'Temperature',
                    'Readable Thresholds': 'unc ucr unr',
                    'Negative Hysteresis': 'Unspecified',
                    'Threshold Read Mask': 'unc ucr unr',
                    'Upper critical': '80.000',
                    'Sensor ID': 'PCI Riser 1 Temp (0x33)',
                    'Settable Thresholds': '',
                    'Minimum sensor range': 'Unspecified',
                    'Nominal Reading': '16.000'
                },
            }
        }
    }
}


BAD_SENSOR = {
    'message_id': 'f22188ca-c068-47ce-a3e5-0e27ffe234c6',
    'publisher_id': 'f23188ca-c068-47ce-a3e5-0e27ffe234c6',
    'payload': {
        'instance_uuid': 'f11251ax-c568-25ca-4582-0x27add644c6',
        'timestamp': '20140223134852',
        'node_uuid': 'f4982fd2-2f2b-4bb5-9aff-48aac801d1ad',
        'event_type': 'hardware.ipmi.metrics.update',
        'payload': {
            'Temperature': {
                'PCI Riser 1 Temp (0x33)': {
                    'Status': 'ok',
                    'Deassertions Enabled': 'unc+ ucr+ unr+',
                    'Sensor Reading': 'some bad stuff',
                    'Entity ID': '16.1 (System Internal Expansion Board)',
                    'Assertions Enabled': 'unc+ ucr+ unr+',
                    'Positive Hysteresis': '4.000',
                    'Assertion Events': '',
                    'Upper non-critical': '70.000',
                    'Event Message Control': 'Per-threshold',
                    'Upper non-recoverable': '85.000',
                    'Normal Maximum': '112.000',
                    'Maximum sensor range': 'Unspecified',
                    'Sensor Type (Analog)': 'Temperature',
                    'Readable Thresholds': 'unc ucr unr',
                    'Negative Hysteresis': 'Unspecified',
                    'Threshold Read Mask': 'unc ucr unr',
                    'Upper critical': '80.000',
                    'Sensor ID': 'PCI Riser 1 Temp (0x33)',
                    'Settable Thresholds': '',
                    'Minimum sensor range': 'Unspecified',
                    'Nominal Reading': '16.000'
                },
            }
        }
    }
}


NO_SENSOR_ID = {
    'message_id': 'f22188ca-c068-47ce-a3e5-0e27ffe234c6',
    'publisher_id': 'f23188ca-c068-47ce-a3e5-0e27ffe234c6',
    'payload': {
        'instance_uuid': 'f11251ax-c568-25ca-4582-0x27add644c6',
        'timestamp': '20140223134852',
        'node_uuid': 'f4982fd2-2f2b-4bb5-9aff-48aac801d1ad',
        'event_type': 'hardware.ipmi.metrics.update',
        'payload': {
            'Temperature': {
                'PCI Riser 1 Temp (0x33)': {
                    'Sensor Reading': '26 C',
                },
            }
        }
    }
}


NO_NODE_ID = {
    'message_id': 'f22188ca-c068-47ce-a3e5-0e27ffe234c6',
    'publisher_id': 'f23188ca-c068-47ce-a3e5-0e27ffe234c6',
    'payload': {
        'instance_uuid': 'f11251ax-c568-25ca-4582-0x27add644c6',
        'timestamp': '20140223134852',
        'event_type': 'hardware.ipmi.metrics.update',
        'payload': {
            'Temperature': {
                'PCI Riser 1 Temp (0x33)': {
                    'Sensor Reading': '26 C',
                    'Sensor ID': 'PCI Riser 1 Temp (0x33)',
                },
            }
        }
    }
}



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import operator
import time

from oslo_log import log
from oslo_utils import timeutils

import ceilometer
from ceilometer.storage import base
from ceilometer.storage.hbase import base as hbase_base
from ceilometer.storage.hbase import migration as hbase_migration
from ceilometer.storage.hbase import utils as hbase_utils
from ceilometer.storage import models
from ceilometer import utils

LOG = log.getLogger(__name__)


AVAILABLE_CAPABILITIES = {
    'meters': {'query': {'simple': True,
                         'metadata': True}},
    'resources': {'query': {'simple': True,
                            'metadata': True}},
    'samples': {'query': {'simple': True,
                          'metadata': True}},
    'statistics': {'query': {'simple': True,
                             'metadata': True},
                   'aggregation': {'standard': True}},
}


AVAILABLE_STORAGE_CAPABILITIES = {
    'storage': {'production_ready': True},
}


class Connection(hbase_base.Connection, base.Connection):
    """Put the metering data into a HBase database

    Collections:

    - meter (describes sample actually):

      - row-key: consists of reversed timestamp, meter and a message uuid
        for purposes of uniqueness
      - Column Families:

        f: contains the following qualifiers:

          - counter_name: <name of counter>
          - counter_type: <type of counter>
          - counter_unit: <unit of counter>
          - counter_volume: <volume of counter>
          - message: <raw incoming data>
          - message_id: <id of message>
          - message_signature: <signature of message>
          - resource_metadata: raw metadata for corresponding resource
            of the meter
          - project_id: <id of project>
          - resource_id: <id of resource>
          - user_id: <id of user>
          - recorded_at: <datetime when sample has been recorded (utc.now)>
          - flattened metadata with prefix r_metadata. e.g.::

             f:r_metadata.display_name or f:r_metadata.tag

          - rts: <reversed timestamp of entry>
          - timestamp: <meter's timestamp (came from message)>
          - source for meter with prefix 's'

    - resource:

      - row_key: uuid of resource
      - Column Families:

        f: contains the following qualifiers:

          - resource_metadata: raw metadata for corresponding resource
          - project_id: <id of project>
          - resource_id: <id of resource>
          - user_id: <id of user>
          - flattened metadata with prefix r_metadata. e.g.::

             f:r_metadata.display_name or f:r_metadata.tag

          - sources for all corresponding meters with prefix 's'
          - all meters with prefix 'm' for this resource in format:

            .. code-block:: python

              "%s:%s:%s:%s:%s" % (rts, source, counter_name, counter_type,
              counter_unit)
    """

    CAPABILITIES = utils.update_nested(base.Connection.CAPABILITIES,
                                       AVAILABLE_CAPABILITIES)
    STORAGE_CAPABILITIES = utils.update_nested(
        base.Connection.STORAGE_CAPABILITIES,
        AVAILABLE_STORAGE_CAPABILITIES,
    )
    _memory_instance = None

    RESOURCE_TABLE = "resource"
    METER_TABLE = "meter"

    def __init__(self, url):
        super(Connection, self).__init__(url)

    def upgrade(self):
        tables = [self.RESOURCE_TABLE, self.METER_TABLE]
        column_families = {'f': dict(max_versions=1)}
        with self.conn_pool.connection() as conn:
            hbase_utils.create_tables(conn, tables, column_families)
            hbase_migration.migrate_tables(conn, tables)

    def clear(self):
        LOG.debug('Dropping HBase schema...')
        with self.conn_pool.connection() as conn:
            for table in [self.RESOURCE_TABLE,
                          self.METER_TABLE]:
                try:
                    conn.disable_table(table)
                except Exception:
                    LOG.debug('Cannot disable table but ignoring error')
                try:
                    conn.delete_table(table)
                except Exception:
                    LOG.debug('Cannot delete table but ignoring error')

    def record_metering_data(self, data):
        """Write the data to the backend storage system.

        :param data: a dictionary such as returned by
          ceilometer.meter.meter_message_from_counter
        """
        with self.conn_pool.connection() as conn:
            resource_table = conn.table(self.RESOURCE_TABLE)
            meter_table = conn.table(self.METER_TABLE)

            resource_metadata = data.get('resource_metadata', {})
            # Determine the name of new meter
            rts = hbase_utils.timestamp(data['timestamp'])
            new_meter = hbase_utils.prepare_key(
                rts, data['source'], data['counter_name'],
                data['counter_type'], data['counter_unit'])

            # TODO(nprivalova): try not to store resource_id
            resource = hbase_utils.serialize_entry(**{
                'source': data['source'],
                'meter': {new_meter: data['timestamp']},
                'resource_metadata': resource_metadata,
                'resource_id': data['resource_id'],
                'project_id': data['project_id'], 'user_id': data['user_id']})
            # Here we put entry in HBase with our own timestamp. This is needed
            # when samples arrive out-of-order
            # If we use timestamp=data['timestamp'] the newest data will be
            # automatically 'on the top'. It is needed to keep metadata
            # up-to-date: metadata from newest samples is considered as actual.
            ts = int(time.mktime(data['timestamp'].timetuple()) * 1000)
            resource_table.put(hbase_utils.encode_unicode(data['resource_id']),
                               resource, ts)

            # Rowkey consists of reversed timestamp, meter and a
            # message uuid for purposes of uniqueness
            row = hbase_utils.prepare_key(data['counter_name'], rts,
                                          data['message_id'])
            record = hbase_utils.serialize_entry(
                data, **{'source': data['source'], 'rts': rts,
                         'message': data, 'recorded_at': timeutils.utcnow()})
            meter_table.put(row, record)

    def get_resources(self, user=None, project=None, source=None,
                      start_timestamp=None, start_timestamp_op=None,
                      end_timestamp=None, end_timestamp_op=None,
                      metaquery=None, resource=None, limit=None):
        """Return an iterable of models.Resource instances

        :param user: Optional ID for user that owns the resource.
        :param project: Optional ID for project that owns the resource.
        :param source: Optional source filter.
        :param start_timestamp: Optional modified timestamp start range.
        :param start_timestamp_op: Optional start time operator, like ge, gt.
        :param end_timestamp: Optional modified timestamp end range.
        :param end_timestamp_op: Optional end time operator, like lt, le.
        :param metaquery: Optional dict with metadata to match on.
        :param resource: Optional resource filter.
        :param limit: Maximum number of results to return.
        """
        if limit == 0:
            return
        q = hbase_utils.make_query(metaquery=metaquery, user_id=user,
                                   project_id=project,
                                   resource_id=resource, source=source)
        q = hbase_utils.make_meter_query_for_resource(start_timestamp,
                                                      start_timestamp_op,
                                                      end_timestamp,
                                                      end_timestamp_op,
                                                      source, q)
        with self.conn_pool.connection() as conn:
            resource_table = conn.table(self.RESOURCE_TABLE)
            LOG.debug("Query Resource table: %s", q)
            for resource_id, data in resource_table.scan(filter=q,
                                                         limit=limit):
                f_res, meters, md = hbase_utils.deserialize_entry(
                    data)
                resource_id = hbase_utils.encode_unicode(resource_id)
                # Unfortunately happybase doesn't keep ordered result from
                # HBase. So that's why it's needed to find min and max
                # manually
                first_ts = min(meters, key=operator.itemgetter(1))[1]
                last_ts = max(meters, key=operator.itemgetter(1))[1]
                source = meters[0][0][1]
                # If we use QualifierFilter then HBase returnes only
                # qualifiers filtered by. It will not return the whole entry.
                # That's why if we need to ask additional qualifiers manually.
                if 'project_id' not in f_res and 'user_id' not in f_res:
                    row = resource_table.row(
                        resource_id, columns=['f:project_id', 'f:user_id',
                                              'f:resource_metadata'])
                    f_res, _m, md = hbase_utils.deserialize_entry(row)
                yield models.Resource(
                    resource_id=resource_id,
                    first_sample_timestamp=first_ts,
                    last_sample_timestamp=last_ts,
                    project_id=f_res['project_id'],
                    source=source,
                    user_id=f_res['user_id'],
                    metadata=md)

    def get_meters(self, user=None, project=None, resource=None, source=None,
                   metaquery=None, limit=None, unique=False):
        """Return an iterable of models.Meter instances

        :param user: Optional ID for user that owns the resource.
        :param project: Optional ID for project that owns the resource.
        :param resource: Optional resource filter.
        :param source: Optional source filter.
        :param metaquery: Optional dict with metadata to match on.
        :param limit: Maximum number of results to return.
        :param unique: If set to true, return only unique meter information.
        """
        if limit == 0:
            return

        metaquery = metaquery or {}

        with self.conn_pool.connection() as conn:
            resource_table = conn.table(self.RESOURCE_TABLE)
            q = hbase_utils.make_query(metaquery=metaquery, user_id=user,
                                       project_id=project,
                                       resource_id=resource,
                                       source=source)
            LOG.debug("Query Resource table: %s", q)

            gen = resource_table.scan(filter=q)
            # We need result set to be sure that user doesn't receive several
            # same meters. Please see bug
            # https://bugs.launchpad.net/ceilometer/+bug/1301371
            result = set()
            for ignored, data in gen:
                flatten_result, meters, md = hbase_utils.deserialize_entry(
                    data)
                for m in meters:
                    if limit and len(result) >= limit:
                        return
                    _m_rts, m_source, name, m_type, unit = m[0]
                    if unique:
                        meter_dict = {'name': name,
                                      'type': m_type,
                                      'unit': unit,
                                      'resource_id': None,
                                      'project_id': None,
                                      'user_id': None,
                                      'source': None}
                    else:
                        meter_dict = {'name': name,
                                      'type': m_type,
                                      'unit': unit,
                                      'resource_id':
                                          flatten_result['resource_id'],
                                      'project_id':
                                          flatten_result['project_id'],
                                      'user_id':
                                          flatten_result['user_id']}

                    frozen_meter = frozenset(meter_dict.items())
                    if frozen_meter in result:
                        continue
                    result.add(frozen_meter)
                    if not unique:
                        meter_dict.update({'source': m_source
                                           if m_source else None})

                    yield models.Meter(**meter_dict)

    def get_samples(self, sample_filter, limit=None):
        """Return an iterable of models.Sample instances.

        :param sample_filter: Filter.
        :param limit: Maximum number of results to return.
        """
        if limit == 0:
            return
        with self.conn_pool.connection() as conn:
            meter_table = conn.table(self.METER_TABLE)
            q, start, stop, columns = (hbase_utils.
                                       make_sample_query_from_filter
                                       (sample_filter, require_meter=False))
            LOG.debug("Query Meter Table: %s", q)
            gen = meter_table.scan(filter=q, row_start=start, row_stop=stop,
                                   limit=limit, columns=columns)
            for ignored, meter in gen:
                d_meter = hbase_utils.deserialize_entry(meter)[0]
                d_meter['message']['counter_volume'] = (
                    float(d_meter['message']['counter_volume']))
                d_meter['message']['recorded_at'] = d_meter['recorded_at']
                yield models.Sample(**d_meter['message'])

    @staticmethod
    def _update_meter_stats(stat, meter):
        """Do the stats calculation on a requested time bucket in stats dict

        :param stats: dict where aggregated stats are kept
        :param index: time bucket index in stats
        :param meter: meter record as returned from HBase
        :param start_time: query start time
        :param period: length of the time bucket
        """
        vol = meter['counter_volume']
        ts = meter['timestamp']
        stat.unit = meter['counter_unit']
        stat.min = min(vol, stat.min or vol)
        stat.max = max(vol, stat.max)
        stat.sum = vol + (stat.sum or 0)
        stat.count += 1
        stat.avg = (stat.sum / float(stat.count))
        stat.duration_start = min(ts, stat.duration_start or ts)
        stat.duration_end = max(ts, stat.duration_end or ts)
        stat.duration = (timeutils.delta_seconds(stat.duration_start,
                                                 stat.duration_end))

    def get_meter_statistics(self, sample_filter, period=None, groupby=None,
                             aggregate=None):
        """Return an iterable of models.Statistics instances.

        Items are containing meter statistics described by the query
        parameters. The filter must have a meter value set.

        .. note::

          Due to HBase limitations the aggregations are implemented
          in the driver itself, therefore this method will be quite slow
          because of all the Thrift traffic it is going to create.
        """
        if groupby:
            raise ceilometer.NotImplementedError("Group by not implemented.")

        if aggregate:
            raise ceilometer.NotImplementedError(
                'Selectable aggregates not implemented')

        with self.conn_pool.connection() as conn:
            meter_table = conn.table(self.METER_TABLE)
            q, start, stop, columns = (hbase_utils.
                                       make_sample_query_from_filter
                                       (sample_filter))
            # These fields are used in statistics' calculating
            columns.extend(['f:timestamp', 'f:counter_volume',
                            'f:counter_unit'])
            meters = map(hbase_utils.deserialize_entry,
                         list(meter for (ignored, meter) in
                              meter_table.scan(
                                  filter=q, row_start=start,
                                  row_stop=stop, columns=columns)))

        if sample_filter.start_timestamp:
            start_time = sample_filter.start_timestamp
        elif meters:
            start_time = meters[-1][0]['timestamp']
        else:
            start_time = None

        if sample_filter.end_timestamp:
            end_time = sample_filter.end_timestamp
        elif meters:
            end_time = meters[0][0]['timestamp']
        else:
            end_time = None

        results = []

        if not period:
            period = 0
            period_start = start_time
            period_end = end_time

        # As our HBase meters are stored as newest-first, we need to iterate
        # in the reverse order
        for meter in meters[::-1]:
            ts = meter[0]['timestamp']
            if period:
                offset = int(timeutils.delta_seconds(
                    start_time, ts) / period) * period
                period_start = start_time + datetime.timedelta(0, offset)

            if not results or not results[-1].period_start == period_start:
                if period:
                    period_end = period_start + datetime.timedelta(
                        0, period)
                results.append(
                    models.Statistics(unit='',
                                      count=0,
                                      min=0,
                                      max=0,
                                      avg=0,
                                      sum=0,
                                      period=period,
                                      period_start=period_start,
                                      period_end=period_end,
                                      duration=None,
                                      duration_start=None,
                                      duration_end=None,
                                      groupby=None)
                )
            self._update_meter_stats(results[-1], meter[0])
        return results



#
# Copyright 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_utils import timeutils

from ceilometer.agent import plugin_base
from ceilometer import nova_client

OPTS = [
    cfg.BoolOpt('workload_partitioning',
                default=False,
                help='Enable work-load partitioning, allowing multiple '
                     'compute agents to be run simultaneously.'),
    cfg.IntOpt('resource_update_interval',
               default=0,
               min=0,
               help="New instances will be discovered periodically based"
                    " on this option (in seconds). By default, "
                    "the agent discovers instances according to pipeline "
                    "polling interval. If option is greater than 0, "
                    "the instance list to poll will be updated based "
                    "on this option's interval. Measurements relating "
                    "to the instances will match intervals "
                    "defined in pipeline.")
]
cfg.CONF.register_opts(OPTS, group='compute')


class InstanceDiscovery(plugin_base.DiscoveryBase):
    def __init__(self):
        super(InstanceDiscovery, self).__init__()
        self.nova_cli = nova_client.Client()
        self.last_run = None
        self.instances = {}
        self.expiration_time = cfg.CONF.compute.resource_update_interval

    def discover(self, manager, param=None):
        """Discover resources to monitor."""
        secs_from_last_update = 0
        if self.last_run:
            secs_from_last_update = timeutils.delta_seconds(
                self.last_run, timeutils.utcnow(True))

        instances = []
        # NOTE(ityaptin) we update make a nova request only if
        # it's a first discovery or resources expired
        if not self.last_run or secs_from_last_update >= self.expiration_time:
            try:
                utc_now = timeutils.utcnow(True)
                since = self.last_run.isoformat() if self.last_run else None
                instances = self.nova_cli.instance_get_all_by_host(
                    cfg.CONF.host, since)
                self.last_run = utc_now
            except Exception:
                # NOTE(zqfan): instance_get_all_by_host is wrapped and will log
                # exception when there is any error. It is no need to raise it
                # again and print one more time.
                return []

        for instance in instances:
            if getattr(instance, 'OS-EXT-STS:vm_state', None) in ['deleted',
                                                                  'error']:
                self.instances.pop(instance.id, None)
            else:
                self.instances[instance.id] = instance

        return self.instances.values()

    @property
    def group_id(self):
        if cfg.CONF.compute.workload_partitioning:
            return cfg.CONF.host
        else:
            return None



#
# Copyright 2012 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Test listing raw events.
"""

import datetime

import mock
from oslo_utils import timeutils

from ceilometer.storage import models
from ceilometer.tests.functional.api import v2


class TestComputeDurationByResource(v2.FunctionalTest):

    def setUp(self):
        super(TestComputeDurationByResource, self).setUp()
        # Create events relative to the range and pretend
        # that the intervening events exist.

        self.early1 = datetime.datetime(2012, 8, 27, 7, 0)
        self.early2 = datetime.datetime(2012, 8, 27, 17, 0)

        self.start = datetime.datetime(2012, 8, 28, 0, 0)

        self.middle1 = datetime.datetime(2012, 8, 28, 8, 0)
        self.middle2 = datetime.datetime(2012, 8, 28, 18, 0)

        self.end = datetime.datetime(2012, 8, 28, 23, 59)

        self.late1 = datetime.datetime(2012, 8, 29, 9, 0)
        self.late2 = datetime.datetime(2012, 8, 29, 19, 0)

    def _patch_get_interval(self, start, end):
        def get_interval(sample_filter, period, groupby, aggregate):
            self.assertIsNotNone(sample_filter.start_timestamp)
            self.assertIsNotNone(sample_filter.end_timestamp)
            if (sample_filter.start_timestamp > end or
                    sample_filter.end_timestamp < start):
                return []
            duration_start = max(sample_filter.start_timestamp, start)
            duration_end = min(sample_filter.end_timestamp, end)
            duration = timeutils.delta_seconds(duration_start, duration_end)
            return [
                models.Statistics(
                    unit='',
                    min=0,
                    max=0,
                    avg=0,
                    sum=0,
                    count=0,
                    period=None,
                    period_start=None,
                    period_end=None,
                    duration=duration,
                    duration_start=duration_start,
                    duration_end=duration_end,
                    groupby=None,
                )
            ]
        return mock.patch.object(type(self.conn), 'get_meter_statistics',
                                 side_effect=get_interval)

    def _invoke_api(self):
        return self.get_json('/meters/instance/statistics',
                             q=[{'field': 'timestamp',
                                 'op': 'ge',
                                 'value': self.start.isoformat()},
                                {'field': 'timestamp',
                                 'op': 'le',
                                 'value': self.end.isoformat()},
                                {'field': 'search_offset',
                                 'value': 10}])

    def test_before_range(self):
        with self._patch_get_interval(self.early1, self.early2):
            data = self._invoke_api()
        self.assertEqual([], data)

    def _assert_times_match(self, actual, expected):
        if actual:
            actual = timeutils.parse_isotime(actual)
        actual = actual.replace(tzinfo=None)
        self.assertEqual(expected, actual)

    def test_overlap_range_start(self):
        with self._patch_get_interval(self.early1, self.middle1):
            data = self._invoke_api()
        self._assert_times_match(data[0]['duration_start'], self.start)
        self._assert_times_match(data[0]['duration_end'], self.middle1)
        self.assertEqual(8 * 60 * 60, data[0]['duration'])

    def test_within_range(self):
        with self._patch_get_interval(self.middle1, self.middle2):
            data = self._invoke_api()
        self._assert_times_match(data[0]['duration_start'], self.middle1)
        self._assert_times_match(data[0]['duration_end'], self.middle2)
        self.assertEqual(10 * 60 * 60, data[0]['duration'])

    def test_within_range_zero_duration(self):
        with self._patch_get_interval(self.middle1, self.middle1):
            data = self._invoke_api()
        self._assert_times_match(data[0]['duration_start'], self.middle1)
        self._assert_times_match(data[0]['duration_end'], self.middle1)
        self.assertEqual(0, data[0]['duration'])

    def test_overlap_range_end(self):
        with self._patch_get_interval(self.middle2, self.late1):
            data = self._invoke_api()
        self._assert_times_match(data[0]['duration_start'], self.middle2)
        self._assert_times_match(data[0]['duration_end'], self.end)
        self.assertEqual(((6 * 60) - 1) * 60, data[0]['duration'])

    def test_after_range(self):
        with self._patch_get_interval(self.late1, self.late2):
            data = self._invoke_api()
        self.assertEqual([], data)

    def test_without_end_timestamp(self):
        statistics = [
            models.Statistics(
                unit=None,
                count=0,
                min=None,
                max=None,
                avg=None,
                duration=None,
                duration_start=self.late1,
                duration_end=self.late2,
                sum=0,
                period=None,
                period_start=None,
                period_end=None,
                groupby=None,
            )
        ]
        with mock.patch.object(type(self.conn), 'get_meter_statistics',
                               return_value=statistics):
            data = self.get_json('/meters/instance/statistics',
                                 q=[{'field': 'timestamp',
                                     'op': 'ge',
                                     'value': self.late1.isoformat()},
                                    {'field': 'resource_id',
                                     'value': 'resource-id'},
                                    {'field': 'search_offset',
                                     'value': 10}])
        self._assert_times_match(data[0]['duration_start'], self.late1)
        self._assert_times_match(data[0]['duration_end'], self.late2)

    def test_without_start_timestamp(self):
        statistics = [
            models.Statistics(
                unit=None,
                count=0,
                min=None,
                max=None,
                avg=None,
                duration=None,
                duration_start=self.early1,
                duration_end=self.early2,
                sum=0,
                period=None,
                period_start=None,
                period_end=None,
                groupby=None,
            )
        ]

        with mock.patch.object(type(self.conn), 'get_meter_statistics',
                               return_value=statistics):
            data = self.get_json('/meters/instance/statistics',
                                 q=[{'field': 'timestamp',
                                     'op': 'le',
                                     'value': self.early2.isoformat()},
                                    {'field': 'resource_id',
                                     'value': 'resource-id'},
                                    {'field': 'search_offset',
                                     'value': 10}])
        self._assert_times_match(data[0]['duration_start'], self.early1)
        self._assert_times_match(data[0]['duration_end'], self.early2)



#
# Copyright 2012 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Handler for producing network counter messages from Neutron notification
   events.

"""

from oslo_config import cfg
import oslo_messaging

from ceilometer.agent import plugin_base
from ceilometer import sample

OPTS = [
    cfg.StrOpt('neutron_control_exchange',
               default='neutron',
               help="Exchange name for Neutron notifications."),
]

cfg.CONF.register_opts(OPTS)


class NetworkNotificationBase(plugin_base.NotificationBase):

    resource_name = None

    @property
    def event_types(self):
        return [
            # NOTE(flwang): When the *.create.start notification sending,
            # there is no resource id assigned by Neutron yet. So we ignore
            # the *.create.start notification for now and only listen the
            # *.create.end to make sure the resource id is existed.
            '%s.create.end' % self.resource_name,
            '%s.update.*' % self.resource_name,
            '%s.exists' % self.resource_name,
            # FIXME(dhellmann): Neutron delete notifications do
            # not include the same metadata as the other messages,
            # so we ignore them for now. This isn't ideal, since
            # it may mean we miss charging for some amount of time,
            # but it is better than throwing away the existing
            # metadata for a resource when it is deleted.
            # '%s.delete.start' % (self.resource_name),
        ]

    def get_targets(self, conf):
        """Return a sequence of oslo_messaging.Target

        This sequence is defining the exchange and topics to be connected for
        this plugin.
        """
        return [oslo_messaging.Target(topic=topic,
                                      exchange=conf.neutron_control_exchange)
                for topic in self.get_notification_topics(conf)]

    def process_notification(self, message):
        counter_name = getattr(self, 'counter_name', self.resource_name)
        unit_value = getattr(self, 'unit', self.resource_name)

        resource = message['payload'].get(self.resource_name)
        if resource:
            # NOTE(liusheng): In %s.update.start notifications, the id is in
            # message['payload'] instead of resource itself.
            if message['event_type'].endswith('update.start'):
                resource['id'] = message['payload']['id']
            resources = [resource]
        else:
            resources = message['payload'].get(self.resource_name + 's', [])

        resource_message = message.copy()
        for resource in resources:
            resource_message['payload'] = resource
            yield sample.Sample.from_notification(
                name=counter_name,
                type=sample.TYPE_GAUGE,
                unit=unit_value,
                volume=1,
                user_id=resource_message['_context_user_id'],
                project_id=resource_message['_context_tenant_id'],
                resource_id=resource['id'],
                message=resource_message)
            event_type_split = resource_message['event_type'].split('.')
            if len(event_type_split) > 2:
                yield sample.Sample.from_notification(
                    name=counter_name
                    + "." + event_type_split[1],
                    type=sample.TYPE_DELTA,
                    unit=unit_value,
                    volume=1,
                    user_id=resource_message['_context_user_id'],
                    project_id=resource_message['_context_tenant_id'],
                    resource_id=resource['id'],
                    message=resource_message)


class Network(NetworkNotificationBase, plugin_base.NonMetricNotificationBase):
    """Listen for Neutron network notifications.

    Handle network.{create.end|update.*|exists} notifications from neutron.
    """
    resource_name = 'network'


class Subnet(NetworkNotificationBase, plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle subnet.{create.end|update.*|exists} notifications from neutron.
    """
    resource_name = 'subnet'


class Port(NetworkNotificationBase, plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle port.{create.end|update.*|exists} notifications from neutron.
    """
    resource_name = 'port'


class Router(NetworkNotificationBase, plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle router.{create.end|update.*|exists} notifications from neutron.
    """
    resource_name = 'router'


class FloatingIP(NetworkNotificationBase,
                 plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle floatingip.{create.end|update.*|exists} notifications from neutron.
    """
    resource_name = 'floatingip'
    counter_name = 'ip.floating'
    unit = 'ip'


class Pool(NetworkNotificationBase, plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle pool.{create.end|update.*|exists} notifications from neutron.
    """
    resource_name = 'pool'
    counter_name = 'network.services.lb.pool'


class Vip(NetworkNotificationBase, plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle vip.{create.end|update.*|exists} notifications from neutron.
    """
    resource_name = 'vip'
    counter_name = 'network.services.lb.vip'


class Member(NetworkNotificationBase, plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle member.{create.end|update.*|exists} notifications from neutron.
    """
    resource_name = 'member'
    counter_name = 'network.services.lb.member'


class HealthMonitor(NetworkNotificationBase,
                    plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle health_monitor.{create.end|update.*|exists} notifications
    from neutron.
    """
    resource_name = 'health_monitor'
    counter_name = 'network.services.lb.health_monitor'


class Firewall(NetworkNotificationBase, plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle firewall.{create.end|update.*|exists} notifications from neutron.
    """
    resource_name = 'firewall'
    counter_name = 'network.services.firewall'


class FirewallPolicy(NetworkNotificationBase,
                     plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle firewall_policy.{create.end|update.*|exists} notifications
    from neutron.
    """
    resource_name = 'firewall_policy'
    counter_name = 'network.services.firewall.policy'


class FirewallRule(NetworkNotificationBase,
                   plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle firewall_rule.{create.end|update.*|exists} notifications
    from neutron.
    """
    resource_name = 'firewall_rule'
    counter_name = 'network.services.firewall.rule'


class VPNService(NetworkNotificationBase,
                 plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle vpnservice.{create.end|update.*|exists} notifications from neutron.
    """
    resource_name = 'vpnservice'
    counter_name = 'network.services.vpn'


class IPSecPolicy(NetworkNotificationBase,
                  plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle pool.{create.end|update.*|exists} notifications from neutron.
    """
    resource_name = 'ipsecpolicy'
    counter_name = 'network.services.vpn.ipsecpolicy'


class IKEPolicy(NetworkNotificationBase,
                plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle ikepolicy.{create.end|update.*|exists} notifications from neutron.
    """
    resource_name = 'ikepolicy'
    counter_name = 'network.services.vpn.ikepolicy'


class IPSecSiteConnection(NetworkNotificationBase,
                          plugin_base.NonMetricNotificationBase):
    """Listen for Neutron notifications.

    Handle ipsec_site_connection.{create.end|update.*|exists}
    notifications from neutron.
    """
    resource_name = 'ipsec_site_connection'
    counter_name = 'network.services.vpn.connections'



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import ceilometer


class Connection(object):
    """Base class for event storage system connections."""

    # A dictionary representing the capabilities of this driver.
    CAPABILITIES = {
        'events': {'query': {'simple': False}},
    }

    STORAGE_CAPABILITIES = {
        'storage': {'production_ready': False},
    }

    def __init__(self, url):
        pass

    @staticmethod
    def upgrade():
        """Migrate the database to `version` or the most recent version."""

    @staticmethod
    def clear():
        """Clear database."""

    @staticmethod
    def record_events(events):
        """Write the events to the backend storage system.

        :param events: a list of model.Event objects.
        """
        raise ceilometer.NotImplementedError('Events not implemented.')

    @staticmethod
    def get_events(event_filter, limit=None):
        """Return an iterable of model.Event objects."""
        raise ceilometer.NotImplementedError('Events not implemented.')

    @staticmethod
    def get_event_types():
        """Return all event types as an iterable of strings."""
        raise ceilometer.NotImplementedError('Events not implemented.')

    @staticmethod
    def get_trait_types(event_type):
        """Return a dictionary containing the name and data type of the trait.

        Only trait types for the provided event_type are
        returned.
        :param event_type: the type of the Event
        """
        raise ceilometer.NotImplementedError('Events not implemented.')

    @staticmethod
    def get_traits(event_type, trait_type=None):
        """Return all trait instances associated with an event_type.

        If trait_type is specified, only return instances of that trait type.
        :param event_type: the type of the Event to filter by
        :param trait_type: the name of the Trait to filter by
        """

        raise ceilometer.NotImplementedError('Events not implemented.')

    @classmethod
    def get_capabilities(cls):
        """Return an dictionary with the capabilities of each driver."""
        return cls.CAPABILITIES

    @classmethod
    def get_storage_capabilities(cls):
        """Return a dictionary representing the performance capabilities.

        This is needed to evaluate the performance of each driver.
        """
        return cls.STORAGE_CAPABILITIES

    @staticmethod
    def clear_expired_event_data(ttl):
        """Clear expired data from the backend storage system.

        Clearing occurs according to the time-to-live.

        :param ttl: Number of seconds to keep records for.
        """
        raise ceilometer.NotImplementedError('Clearing events not implemented')



# Copyright 2014 Intel
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Tests for xenapi inspector.
"""

import mock
from oslotest import base

from ceilometer.compute.virt import inspector as virt_inspector
from ceilometer.compute.virt.xenapi import inspector as xenapi_inspector


class TestSwapXapiHost(base.BaseTestCase):

    def test_swapping(self):
        self.assertEqual(
            "http://otherserver:8765/somepath",
            xenapi_inspector.swap_xapi_host(
                "http://someserver:8765/somepath", 'otherserver'))

    def test_no_port(self):
        self.assertEqual(
            "http://otherserver/somepath",
            xenapi_inspector.swap_xapi_host(
                "http://someserver/somepath", 'otherserver'))

    def test_no_path(self):
        self.assertEqual(
            "http://otherserver",
            xenapi_inspector.swap_xapi_host(
                "http://someserver", 'otherserver'))

    def test_same_hostname_path(self):
        self.assertEqual(
            "http://other:80/some",
            xenapi_inspector.swap_xapi_host(
                "http://some:80/some", 'other'))


class TestXenapiInspection(base.BaseTestCase):

    def setUp(self):
        api_session = mock.Mock()
        xenapi_inspector.get_api_session = mock.Mock(return_value=api_session)
        self.inspector = xenapi_inspector.XenapiInspector()

        super(TestXenapiInspection, self).setUp()

    def test_inspect_cpu_util(self):
        fake_instance = {'OS-EXT-SRV-ATTR:instance_name': 'fake_instance_name',
                         'id': 'fake_instance_id'}
        fake_stat = virt_inspector.CPUUtilStats(util=40)

        def fake_xenapi_request(method, args):
            metrics_rec = {
                'memory_actual': '536870912',
                'VCPUs_number': '1',
                'VCPUs_utilisation': {'0': 0.4, }
            }

            if method == 'VM.get_by_name_label':
                return ['vm_ref']
            elif method == 'VM.get_metrics':
                return 'metrics_ref'
            elif method == 'VM_metrics.get_record':
                return metrics_rec
            else:
                return None

        session = self.inspector.session
        with mock.patch.object(session, 'xenapi_request',
                               side_effect=fake_xenapi_request):
            cpu_util_stat = self.inspector.inspect_cpu_util(fake_instance)
            self.assertEqual(fake_stat, cpu_util_stat)

    def test_inspect_memory_usage(self):
        fake_instance = {'OS-EXT-SRV-ATTR:instance_name': 'fake_instance_name',
                         'id': 'fake_instance_id'}
        fake_stat = virt_inspector.MemoryUsageStats(usage=128)

        def fake_xenapi_request(method, args):
            metrics_rec = {
                'memory_actual': '134217728',
            }

            if method == 'VM.get_by_name_label':
                return ['vm_ref']
            elif method == 'VM.get_metrics':
                return 'metrics_ref'
            elif method == 'VM_metrics.get_record':
                return metrics_rec
            else:
                return None

        session = self.inspector.session
        with mock.patch.object(session, 'xenapi_request',
                               side_effect=fake_xenapi_request):
            memory_stat = self.inspector.inspect_memory_usage(fake_instance)
            self.assertEqual(fake_stat, memory_stat)

    def test_inspect_vnic_rates(self):
        fake_instance = {'OS-EXT-SRV-ATTR:instance_name': 'fake_instance_name',
                         'id': 'fake_instance_id'}

        def fake_xenapi_request(method, args):
            vif_rec = {
                'metrics': 'vif_metrics_ref',
                'uuid': 'vif_uuid',
                'MAC': 'vif_mac',
            }

            vif_metrics_rec = {
                'io_read_kbs': '1',
                'io_write_kbs': '2',
            }
            if method == 'VM.get_by_name_label':
                return ['vm_ref']
            elif method == 'VM.get_VIFs':
                return ['vif_ref']
            elif method == 'VIF.get_record':
                return vif_rec
            elif method == 'VIF.get_metrics':
                return 'vif_metrics_ref'
            elif method == 'VIF_metrics.get_record':
                return vif_metrics_rec
            else:
                return None

        session = self.inspector.session
        with mock.patch.object(session, 'xenapi_request',
                               side_effect=fake_xenapi_request):
            interfaces = list(self.inspector.inspect_vnic_rates(fake_instance))

            self.assertEqual(1, len(interfaces))
            vnic0, info0 = interfaces[0]
            self.assertEqual('vif_uuid', vnic0.name)
            self.assertEqual('vif_mac', vnic0.mac)
            self.assertEqual(1024, info0.rx_bytes_rate)
            self.assertEqual(2048, info0.tx_bytes_rate)

    def test_inspect_disk_rates(self):
        fake_instance = {'OS-EXT-SRV-ATTR:instance_name': 'fake_instance_name',
                         'id': 'fake_instance_id'}

        def fake_xenapi_request(method, args):
            vbd_rec = {
                'device': 'xvdd'
            }

            vbd_metrics_rec = {
                'io_read_kbs': '1',
                'io_write_kbs': '2'
            }
            if method == 'VM.get_by_name_label':
                return ['vm_ref']
            elif method == 'VM.get_VBDs':
                return ['vbd_ref']
            elif method == 'VBD.get_record':
                return vbd_rec
            elif method == 'VBD.get_metrics':
                return 'vbd_metrics_ref'
            elif method == 'VBD_metrics.get_record':
                return vbd_metrics_rec
            else:
                return None

        session = self.inspector.session
        with mock.patch.object(session, 'xenapi_request',
                               side_effect=fake_xenapi_request):
            disks = list(self.inspector.inspect_disk_rates(fake_instance))

            self.assertEqual(1, len(disks))
            disk0, info0 = disks[0]
            self.assertEqual('xvdd', disk0.device)
            self.assertEqual(1024, info0.read_bytes_rate)
            self.assertEqual(2048, info0.write_bytes_rate)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

import happybase


class MockHBaseTable(happybase.Table):

    def __init__(self, name, connection, data_prefix):
        # data_prefix is added to all rows which are written
        # in this test. It allows to divide data from different tests
        self.data_prefix = data_prefix
        # We create happybase Table with prefix from
        # CEILOMETER_TEST_HBASE_TABLE_PREFIX
        prefix = os.getenv("CEILOMETER_TEST_HBASE_TABLE_PREFIX", 'test')
        separator = os.getenv(
            "CEILOMETER_TEST_HBASE_TABLE_PREFIX_SEPARATOR", '_')
        super(MockHBaseTable, self).__init__(
            "%s%s%s" % (prefix, separator, name),
            connection)

    def put(self, row, *args, **kwargs):
        row = self.data_prefix + row
        return super(MockHBaseTable, self).put(row, *args,
                                               **kwargs)

    def scan(self, row_start=None, row_stop=None, row_prefix=None,
             columns=None, filter=None, timestamp=None,
             include_timestamp=False, batch_size=10, scan_batching=None,
             limit=None, sorted_columns=False):
        # Add data prefix for row parameters
        # row_prefix could not be combined with row_start or row_stop
        if not row_start and not row_stop:
            row_prefix = self.data_prefix + (row_prefix or "")
            row_start = None
            row_stop = None
        elif row_start and not row_stop:
            # Adding data_prefix to row_start and row_stop does not work
            # if it looks like row_start = %data_prefix%foo,
            # row_stop = %data_prefix, because row_start > row_stop
            filter = self._update_filter_row(filter)
            row_start = self.data_prefix + row_start
        else:
            row_start = self.data_prefix + (row_start or "")
            row_stop = self.data_prefix + (row_stop or "")
        gen = super(MockHBaseTable, self).scan(row_start, row_stop,
                                               row_prefix, columns,
                                               filter, timestamp,
                                               include_timestamp, batch_size,
                                               scan_batching, limit,
                                               sorted_columns)
        data_prefix_len = len(self.data_prefix)
        # Restore original row format
        for row, data in gen:
            yield (row[data_prefix_len:], data)

    def row(self, row, *args, **kwargs):
        row = self.data_prefix + row
        return super(MockHBaseTable, self).row(row, *args, **kwargs)

    def delete(self, row, *args, **kwargs):
        row = self.data_prefix + row
        return super(MockHBaseTable, self).delete(row, *args, **kwargs)

    def _update_filter_row(self, filter):
        if filter:
            return "PrefixFilter(%s) AND %s" % (self.data_prefix, filter)
        else:
            return "PrefixFilter(%s)" % self.data_prefix



# -*- encoding: utf-8 -*-
#
# Copyright 2014-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log
from oslo_service import service as os_service

from ceilometer.agent import manager
from ceilometer.i18n import _LW
from ceilometer import service

LOG = log.getLogger(__name__)

CONF = cfg.CONF


class MultiChoicesOpt(cfg.Opt):
    def __init__(self, name, choices=None, **kwargs):
        super(MultiChoicesOpt, self).__init__(
            name, type=DeduplicatedCfgList(choices), **kwargs)
        self.choices = choices

    def _get_argparse_kwargs(self, group, **kwargs):
        """Extends the base argparse keyword dict for multi choices options."""
        kwargs = super(MultiChoicesOpt, self)._get_argparse_kwargs(group)
        kwargs['nargs'] = '+'
        choices = kwargs.get('choices', self.choices)
        if choices:
            kwargs['choices'] = choices
        return kwargs


class DeduplicatedCfgList(cfg.types.List):
    def __init__(self, choices=None, **kwargs):
        super(DeduplicatedCfgList, self).__init__(**kwargs)
        self.choices = choices or []

    def __call__(self, *args, **kwargs):
        result = super(DeduplicatedCfgList, self).__call__(*args, **kwargs)
        result_set = set(result)
        if len(result) != len(result_set):
            LOG.warning(_LW("Duplicated values: %s found in CLI options, "
                            "auto de-duplicated"), result)
            result = list(result_set)
        if self.choices and not (result_set <= set(self.choices)):
            raise Exception('Valid values are %s, but found %s'
                            % (self.choices, result))
        return result


CLI_OPTS = [
    MultiChoicesOpt('polling-namespaces',
                    default=['compute', 'central'],
                    choices=['compute', 'central', 'ipmi'],
                    dest='polling_namespaces',
                    help='Polling namespace(s) to be used while '
                         'resource polling'),
    MultiChoicesOpt('pollster-list',
                    default=[],
                    dest='pollster_list',
                    help='List of pollsters (or wildcard templates) to be '
                         'used while polling'),
]

CONF.register_cli_opts(CLI_OPTS)


def main():
    service.prepare_service()
    os_service.launch(CONF, manager.AgentManager(CONF.polling_namespaces,
                                                 CONF.pollster_list)).wait()



#
# Copyright Ericsson AB 2013. All rights reserved
#
# Authors: Ildiko Vancsa <ildiko.vancsa@ericsson.com>
#          Balazs Gibizer <balazs.gibizer@ericsson.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Test the methods related to complex query."""
import datetime

import fixtures
import jsonschema
import mock
from oslotest import base
import wsme

from ceilometer.api.controllers.v2 import query
from ceilometer.storage import models


class FakeComplexQuery(query.ValidatedComplexQuery):
    def __init__(self, db_model, additional_name_mapping=None, metadata=False):
        super(FakeComplexQuery, self).__init__(query=None,
                                               db_model=db_model,
                                               additional_name_mapping=(
                                                   additional_name_mapping or
                                                   {}),
                                               metadata_allowed=metadata)


sample_name_mapping = {"resource": "resource_id",
                       "meter": "counter_name",
                       "type": "counter_type",
                       "unit": "counter_unit",
                       "volume": "counter_volume"}


class TestComplexQuery(base.BaseTestCase):
    def setUp(self):
        super(TestComplexQuery, self).setUp()
        self.useFixture(fixtures.MonkeyPatch(
            'pecan.response', mock.MagicMock()))
        self.query = FakeComplexQuery(models.Sample,
                                      sample_name_mapping,
                                      True)

    def test_replace_isotime_utc(self):
        filter_expr = {"=": {"timestamp": "2013-12-05T19:38:29Z"}}
        self.query._replace_isotime_with_datetime(filter_expr)
        self.assertEqual(datetime.datetime(2013, 12, 5, 19, 38, 29),
                         filter_expr["="]["timestamp"])

    def test_replace_isotime_timezone_removed(self):
        filter_expr = {"=": {"timestamp": "2013-12-05T20:38:29+01:00"}}
        self.query._replace_isotime_with_datetime(filter_expr)
        self.assertEqual(datetime.datetime(2013, 12, 5, 20, 38, 29),
                         filter_expr["="]["timestamp"])

    def test_replace_isotime_wrong_syntax(self):
        filter_expr = {"=": {"timestamp": "not a valid isotime string"}}
        self.assertRaises(wsme.exc.ClientSideError,
                          self.query._replace_isotime_with_datetime,
                          filter_expr)

    def test_replace_isotime_in_complex_filter(self):
        filter_expr = {"and": [{"=": {"timestamp": "2013-12-05T19:38:29Z"}},
                               {"=": {"timestamp": "2013-12-06T19:38:29Z"}}]}
        self.query._replace_isotime_with_datetime(filter_expr)
        self.assertEqual(datetime.datetime(2013, 12, 5, 19, 38, 29),
                         filter_expr["and"][0]["="]["timestamp"])
        self.assertEqual(datetime.datetime(2013, 12, 6, 19, 38, 29),
                         filter_expr["and"][1]["="]["timestamp"])

    def test_replace_isotime_in_complex_filter_with_unbalanced_tree(self):
        subfilter = {"and": [{"=": {"project_id": 42}},
                             {"=": {"timestamp": "2013-12-06T19:38:29Z"}}]}

        filter_expr = {"or": [{"=": {"timestamp": "2013-12-05T19:38:29Z"}},
                              subfilter]}

        self.query._replace_isotime_with_datetime(filter_expr)
        self.assertEqual(datetime.datetime(2013, 12, 5, 19, 38, 29),
                         filter_expr["or"][0]["="]["timestamp"])
        self.assertEqual(datetime.datetime(2013, 12, 6, 19, 38, 29),
                         filter_expr["or"][1]["and"][1]["="]["timestamp"])

    def test_convert_operator_to_lower_case(self):
        filter_expr = {"AND": [{"=": {"project_id": 42}},
                               {"=": {"project_id": 44}}]}
        self.query._convert_operator_to_lower_case(filter_expr)
        self.assertEqual("and", list(filter_expr.keys())[0])

        filter_expr = {"Or": [{"=": {"project_id": 43}},
                              {"anD": [{"=": {"project_id": 44}},
                                       {"=": {"project_id": 42}}]}]}
        self.query._convert_operator_to_lower_case(filter_expr)
        self.assertEqual("or", list(filter_expr.keys())[0])
        self.assertEqual("and", list(filter_expr["or"][1].keys())[0])

    def test_invalid_filter_misstyped_field_name_samples(self):
        filter = {"=": {"project_id11": 42}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_invalid_complex_filter_wrong_field_names(self):
        filter = {"and":
                  [{"=": {"non_existing_field": 42}},
                   {"=": {"project_id": 42}}]}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

        filter = {"or":
                  [{"=": {"non_existing_field": 42}},
                   {"and":
                    [{"=": {"project_id": 44}},
                     {"=": {"project_id": 42}}]}]}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_convert_orderby(self):
        orderby = []
        self.query._convert_orderby_to_lower_case(orderby)
        self.assertEqual([], orderby)

        orderby = [{"project_id": "DESC"}]
        self.query._convert_orderby_to_lower_case(orderby)
        self.assertEqual([{"project_id": "desc"}], orderby)

        orderby = [{"project_id": "ASC"}, {"resource_id": "DESC"}]
        self.query._convert_orderby_to_lower_case(orderby)
        self.assertEqual([{"project_id": "asc"}, {"resource_id": "desc"}],
                         orderby)

    def test_validate_orderby_empty_direction(self):
        orderby = [{"project_id": ""}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)
        orderby = [{"project_id": "asc"}, {"resource_id": ""}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)

    def test_validate_orderby_wrong_order_string(self):
        orderby = [{"project_id": "not a valid order"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)

    def test_validate_orderby_wrong_multiple_item_order_string(self):
        orderby = [{"project_id": "not a valid order"}, {"resource_id": "ASC"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)

    def test_validate_orderby_empty_field_name(self):
        orderby = [{"": "ASC"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)
        orderby = [{"project_id": "asc"}, {"": "desc"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)

    def test_validate_orderby_wrong_field_name(self):
        orderby = [{"project_id11": "ASC"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)

    def test_validate_orderby_wrong_field_name_multiple_item_orderby(self):
        orderby = [{"project_id": "asc"}, {"resource_id11": "ASC"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)

    def test_validate_orderby_metadata_is_not_allowed(self):
        orderby = [{"metadata.display_name": "asc"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)


class TestFilterSyntaxValidation(base.BaseTestCase):
    def setUp(self):
        super(TestFilterSyntaxValidation, self).setUp()
        self.query = FakeComplexQuery(models.Sample,
                                      sample_name_mapping,
                                      True)

    def test_simple_operator(self):
        filter = {"=": {"project_id": "string_value"}}
        self.query._validate_filter(filter)

        filter = {"=>": {"project_id": "string_value"}}
        self.query._validate_filter(filter)

    def test_valid_value_types(self):
        filter = {"=": {"project_id": "string_value"}}
        self.query._validate_filter(filter)

        filter = {"=": {"project_id": 42}}
        self.query._validate_filter(filter)

        filter = {"=": {"project_id": 3.14}}
        self.query._validate_filter(filter)

        filter = {"=": {"project_id": True}}
        self.query._validate_filter(filter)

        filter = {"=": {"project_id": False}}
        self.query._validate_filter(filter)

    def test_invalid_simple_operator(self):
        filter = {"==": {"project_id": "string_value"}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

        filter = {"": {"project_id": "string_value"}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_more_than_one_operator_is_invalid(self):
        filter = {"=": {"project_id": "string_value"},
                  "<": {"": ""}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_empty_expression_is_invalid(self):
        filter = {}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_invalid_field_name(self):
        filter = {"=": {"": "value"}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

        filter = {"=": {" ": "value"}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

        filter = {"=": {"\t": "value"}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_more_than_one_field_is_invalid(self):
        filter = {"=": {"project_id": "value", "resource_id": "value"}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_missing_field_after_simple_op_is_invalid(self):
        filter = {"=": {}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_and_or(self):
        filter = {"and": [{"=": {"project_id": "string_value"}},
                          {"=": {"resource_id": "value"}}]}
        self.query._validate_filter(filter)

        filter = {"or": [{"and": [{"=": {"project_id": "string_value"}},
                                  {"=": {"resource_id": "value"}}]},
                         {"=": {"counter_name": "value"}}]}
        self.query._validate_filter(filter)

        filter = {"or": [{"and": [{"=": {"project_id": "string_value"}},
                                  {"=": {"resource_id": "value"}},
                                  {"<": {"counter_name": 42}}]},
                         {"=": {"counter_name": "value"}}]}
        self.query._validate_filter(filter)

    def test_complex_operator_with_in(self):
        filter = {"and": [{"<": {"counter_volume": 42}},
                          {">=": {"counter_volume": 36}},
                          {"in": {"project_id": ["project_id1",
                                                 "project_id2",
                                                 "project_id3"]}}]}
        self.query._validate_filter(filter)

    def test_invalid_complex_operator(self):
        filter = {"xor": [{"=": {"project_id": "string_value"}},
                          {"=": {"resource_id": "value"}}]}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_and_or_with_one_child_is_invalid(self):
        filter = {"or": [{"=": {"project_id": "string_value"}}]}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_complex_operator_with_zero_child_is_invalid(self):
        filter = {"or": []}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_more_than_one_complex_operator_is_invalid(self):
        filter = {"and": [{"=": {"project_id": "string_value"}},
                          {"=": {"resource_id": "value"}}],
                  "or": [{"=": {"project_id": "string_value"}},
                         {"=": {"resource_id": "value"}}]}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_not(self):
        filter = {"not": {"=": {"project_id": "value"}}}
        self.query._validate_filter(filter)

        filter = {
            "not":
            {"or":
             [{"and":
               [{"=": {"project_id": "string_value"}},
                {"=": {"resource_id": "value"}},
                {"<": {"counter_name": 42}}]},
              {"=": {"counter_name": "value"}}]}}
        self.query._validate_filter(filter)

    def test_not_with_zero_child_is_invalid(self):
        filter = {"not": {}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_not_with_more_than_one_child_is_invalid(self):
        filter = {"not": {"=": {"project_id": "value"},
                          "!=": {"resource_id": "value"}}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_empty_in_query_not_passing(self):
        filter = {"in": {"resource_id": []}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)
